"""This module exists to support backward compatibility for the expansion of downloader"""

__copyright__ = "Copyright (C) 2019-present, DV Klopfenstein. All rights reserved."
__author__ = "DV Klopfenstein"


# pylint: disable=unused-import
from pmidcite.icite.dnldr.pmid_dnlder import NIHiCiteDownloader


# Copyright (C) 2019-present DV Klopfenstein. All rights reserved.



"""Hold list of Authors."""

__author__ = 'DV Klopfenstein'
__copyright__ = "Copyright (C) 2019-present DV Klopfenstein. All rights reserved."

from pmidcite.eutils.pubmed.author import Author


class Authors(object):
  """Hold list of Authors."""

  flds = ['FAU', 'AU', 'AD']

  def __init__(self):
    self.authors = []
    self.ads = []  # Affiliations for the entire paper

  def get_match(self, mtch_fnc):
    """Find author which matches the search."""
    mtch_idxs = []
    for idx, auobj in enumerate(self.authors):
      if mtch_fnc(auobj):
        mtch_idxs.append(idx)
    return mtch_idxs
    #   auobj = pub['Authors'].get_au_order_desc(au_str)
    #   if au_str == aus[0]:
    #     desc2pub['1st'].append(pub)
    #   elif au_str == aus[-1]:
    #     desc2pub['Last'].append(pub)
    #   else:
    #     desc2pub['mid'].append(pub)

  def add_fld(self, fld, line, pmid):
    """Add an Author or add to the last Author."""
    if fld == 'FAU':  # 'Full Author Name' begins a new Author
      self.authors.append(Author(line))
      return
    # Author affiliation is for all authors on the paper
    if fld == 'AD' and not self.authors:
      self.ads.append(line)
      return
    assert self.authors, "{PMID} {F} {L}".format(PMID=pmid, F=fld, L=line)
    author = self.authors[-1]  # Get the last author
    if fld == 'AU':
      author.au = line
    elif fld == 'AD':
      author.ads.append(line)
    else:
      raise RuntimeError("UNEXPECTED Author FIELD({F})".format(F=fld))

  def __str__(self):
    return "\n".join([str(a) for a in self.authors])


# Copyright (C) 2019-present DV Klopfenstein.  All rights reserved.



"""Write Python module for downloaded abstracts."""

__copyright__ = "Copyright (C) 2019-present, DV Klopfenstein. All rights reserved."
__author__ = "DV Klopfenstein"

import sys
import os
import datetime

from re import match
from re import search
from pmidcite.eutils.pubmed.terms import MeshTerms
from pmidcite.eutils.pubmed.authors import Authors


class PubMedRdWr:
    """Read downloaded PMID text file [FILE].Entrez_downloads. Write subset into Python module."""

    # https://www.nlm.nih.gov/bsd/mms/medlineelements.html
    flds = [
        'TI',  # Title
        'LID', # Location Identifier
        'AID', # Article Identifier
        'PMC', # PubMed Central Identifier
        'OWN', # Owner
        'STATUS', # Status
        'DP',  # Date of Publication
        'AB',  # Abstract
        'JT',  # Journal Title
        'TA',  # Journal Title abbreviation
        'PT',  # Publication Type
        'MH',  # MeSH Terms
        'FAU', # Full Author
        'AU',  # Author
        'AD',  # Affiliation
        'OTO', # Other Term Owner
        'OT',  # Other Term
    ]

    # pylint: disable=bad-continuation
    _lendate2fmt = { # PMID date formats: "2017 Mar 02"
         4 : "%Y",
         8 : "%Y %b",
         9 : "%Y %b.",
        10 : "%Y %b %d",
        11 : "%Y %b %d"}


    status = {
        'In-Data-Review',
        'In-Process',
        'MEDLINE',
        'OLDMEDLINE',
        'PubMed-not-MEDLINE',
        'Publisher',
    }

    owner = {
        'KIE': 'Kennedy Institute of Ethics, Georgetown University',
        'NASA': 'National Aeronautics and Space Administration',
        'NLM': 'National Library of Medicine (used for the OLDMEDLINE records)',
        'NOTNLM': ('(The journal publisher or other data provider. '
                   'This is used when the OT field includes author-supplied keywords. '
                   'NLM began using this value in January 2013.'),
        'PIP': ('Population Information Program, '
                'Johns Hopkins School of Public health '
                '(not a current value; only on older citations)'),
        'HSR': ('National Information Center on '
                'Health Services Research and Health Care Technology, '
                'National Library of Medicine'),
        'HMD': 'History of Medicine Division, National Library of Medicine',
        'SIS': ('Specialized Information Services Division, '
                'National Library of Medicine '
                '(not yet used; reserved for possible future use)'),
    }

    join_these = {
        'TI':' ', # separate by one space when joining
        'DP':'',  # There is only one, but convert from list to str
        'AB':' ',
        'JT':' ',
        'TA':' ',
        'AD':' ', # Author affiliation(s) Each one can be multiple lines
    }

    def __init__(self):
        # These are one block of text, which may be spread over multiple lines
        self.dates = ['DP']
        self.objmh = MeshTerms()
        self.field2fnc = {
            'PMID': self._to_int,
            'MH': self._fld_add_to_list,
            # Indentifiers
            'LID': self._lid_add_to_dict,
            'AID': self._lid_add_to_dict,
            # Date of Publication
            'DP': self._init_date,
            # Author fields
            'FAU': self._add_author,
            'AU': self._add_author,
            'AD': self._add_author
        }

    def _get_fldobjs_all(self, pmid2fldlines):
        """Given a list of lines in a PubMed entry, return objects."""
        pmid2fldobjs = {}
        for pmid, fldlines in pmid2fldlines.items():
            pmid2fldobjs[pmid] = self._get_fldobjs_one(fldlines, pmid)
        return pmid2fldobjs

    def _get_fldobjs_one(self, fldlines, pmid):
        """Given a list of lines in a PubMed entry, return objects."""
        fld2objs = {}
        field2fnc = self.field2fnc
        for fld, line in fldlines:
            # Mesh Terms
            line = " ".join(line)
            if fld in field2fnc:
                field2fnc[fld](fld2objs, fld, line, pmid)
            else:
                fld2objs[fld] = line
            # Author list
            if fld == 'AU':
                self._fld_add_to_list(fld2objs, fld, line, pmid)
        # die
        return fld2objs

    # pylint: disable=unused-argument
    @staticmethod
    def _to_int(fld2objs, fld, line, pmid):
        """Add a value to a list."""
        fld2objs[fld] = int(line)

    @staticmethod
    def _add_author(fld2objs, fld, line, pmid):
        """Add a value to a list."""
        if 'Authors' not in fld2objs:
            fld2objs['Authors'] = Authors()
        fld2objs['Authors'].add_fld(fld, line, pmid)

    @staticmethod
    def _fld_add_to_list(fld2objs, fld, line, pmid):
        """Add a value to a list."""
        if fld not in fld2objs:
            fld2objs[fld] = []
        fld2objs[fld].append(line)

    @staticmethod
    def _lid_add_to_dict(fld2objs, fld, line, pmid):
        """Add a value to a list."""
        if fld not in fld2objs:
            fld2objs[fld] = {}
        key0 = line.rfind('[')
        # TBD Change these fatals to messages
        assert key0 != -1, '**FATAL LID: {} {}'.format(fld, line)
        assert line[-1] == ']', '**FATAL LID: {} {}'.format(fld, line)
        key = line[key0 + 1:-1]
        val = line[:key0].strip()
        assert key not in fld2objs, '**FATAL LID: {} {} {}'.format(key, val, fld2objs)
        fld2objs[fld][key] = val

    def read_pmid2obj_g_fins(self, fins_pubmed, prt=sys.stdout):
        """Load PubMed objects from text files"""
        pmid2obj = {}
        s_get_pmid2info_g_text = self.get_pmid2info_g_text
        for fin_pubmed in fins_pubmed:
            pmid2fld2obj = s_get_pmid2info_g_text(fin_pubmed, prt=prt)
            for pmid, dct in pmid2fld2obj.items():
                pmid2obj[pmid] = dct
        return pmid2obj

    def get_pmid2info_g_text(self, fin_text, flds=None, dochk=False, prt=sys.stdout):
        """Given file of retmode=text, get data (in flds list) as a python dict."""
        if flds is None:
            flds = self.flds
        # Field list: https://www.nlm.nih.gov/bsd/mms/medlineelements.html
        if os.path.exists(fin_text):
            if prt:
                prt.write("  READING: {PUBMED}\n".format(PUBMED=fin_text))
            with open(fin_text) as ifstrm:
                pmid2fldlines = ProcessLines(flds).process_rawlines(ifstrm)
                pmid2fld2objs = self._get_fldobjs_all(pmid2fldlines)
                if prt:
                    prt.write("  READ  {:>7,} PubMed records {} for fields({})\n".format(
                    len(pmid2fld2objs), fin_text, " ".join(flds)))
                if pmid2fld2objs:
                    if dochk:
                        self._chk(pmid2fld2objs)
                return pmid2fld2objs
        else:
            print("  READ  -- 0 -- PubMed record {}".format(fin_text))
            return {}

    def get_pmid2info_g_textblock(self, textblock, flds=None, dochk=False, prt=sys.stdout):
        """Given file of retmode=text, get data (in flds list) as a python dict."""
        if flds is None:
            flds = self.flds
        # Field list: https://www.nlm.nih.gov/bsd/mms/medlineelements.html
        if textblock:
            pmid2fldlines = ProcessLines(flds).process_rawlines(textblock.split('\n'))
            pmid2fld2objs = self._get_fldobjs_all(pmid2fldlines)
            if prt:
                prt.write("  PROCESSED  {P:>7,} PubMed records for fields({Fs})\n".format(
                    P=len(pmid2fld2objs), Fs=" ".join(flds)))
            if pmid2fld2objs:
                if dochk:
                    self._chk(pmid2fld2objs)
            return pmid2fld2objs
        print("  **WARNING: BAD TEXT: {TXT}".format(TXT=textblock))
        return {}

    def _chk(self, pmid2flds):
        """Ensure that MeSH terms are correct."""
        chk_mhs = self.objmh.chk_mhs
        for fld2vals_raw in pmid2flds.values():
            for fld, vals in fld2vals_raw.items():
                if fld == "MH":
                    chk_mhs(vals)

    def wrpy_pmid2flds(self, fout_py, pmid2info, flds=None, pydoc=None):
        """Write PMID info into Python module."""
        import PyBiocode.Utils.dnld as DL
        if flds is None:
            flds = self.flds
        if pydoc is None:
            pydoc = "PMIDs associated with genes in clusters: {FLDS}".format(FLDS=" ".join(flds))
        num_pmids = len(pmid2info)
        with open(fout_py, 'w') as prt:
            prt.write('"""{DESC}"""\n\n'.format(DESC=pydoc))
            prt.write("# {COPYRIGHT}\n\n".format(COPYRIGHT=DL.get_copyright()))
            prt.write("import datetime\n\n")
            prt.write('downloaded = "{DATE}" # {N} items\n\n'.format(
                N=num_pmids, DATE=DL.get_date()))
            prt.write('num_items = {N}\n\n'.format(N=num_pmids))
            prt.write('pmid2info = {\n')
            #### addquot = lambda valstr: ''.join(["'", valstr, "'"])
            for pmid, fld2val_all in self._wrpy_get_sorted(pmid2info, flds):
                fld2val_cur = self._get_fld2val_cur(fld2val_all, flds)
                if fld2val_cur:
                    self._wrpy_pmid(prt, pmid, flds, fld2val_cur)
            prt.write("}\n\n") # End of pmid2info
            prt.write("# {COPYRIGHT}\n\n".format(COPYRIGHT=DL.get_copyright()))
            sys.stdout.write("  WROTE: {}\n".format(fout_py))

    def _wrpy_pmid(self, prt, pmid, flds, fld2val):
        """Write a PMID entry."""
        prt.write("  {PMID:>8} : {{\n".format(PMID=pmid))
        for fld_key in flds:
            fld_val = fld2val.get(fld_key, None)
            if fld_val is not None:
                #### prt.write("    {KEY} : ".format(KEY=addquot(fld_key)))
                prt.write("    {KEY} : ".format(KEY=''.join(["'", fld_key, "'"])))
                # Field: Date
                if fld_key in self.dates:
                    prt.write("{VAL}, # {C}\n".format(
                        VAL=repr(fld_val), C=fld_val.strftime("%Y %b %d")))
                # Field: TI, Abstract, etc.
                elif fld_key in self.join_these.keys():
                    self._wrpyfld_joined(prt, fld_val, pmid)
                # Field: list
                else:
                    prt.write("{VAL},\n".format(VAL=fld_val))
        prt.write("  },\n") # End of dictionary for one PMID

    @staticmethod
    def _get_fld2val_cur(fld2val_all, flds):
        """Get reduced set of fld2val."""
        fld2val_cur = {}
        for fld_key in flds:
            fld_val = fld2val_all.get(fld_key, None)
            if fld_val is not None:
                fld2val_cur[fld_key] = fld_val
        return fld2val_cur

    @staticmethod
    def _wrpy_get_sorted(pmid2info, flds):
        """Return pmid2info sorted by either date or by PubMed ID."""
        if 'DP' in flds:
            return sorted(pmid2info.items(), key=lambda t: t[1]['DP'], reverse=True)
        return sorted(pmid2info.items())

    @staticmethod
    def _wrpyfld_joined(prt, fld_val, pmid):
        """Write field into Python module."""
        #VAL_str = ''.join(['"', fld_val, '"'])
        if "'" in fld_val or '"' in fld_val:
            if '"""' not in fld_val:
                prt.write("{VAL},\n".format(VAL=''.join(['"""', fld_val, ' """'])))
            elif "'''" not in fld_val:
                prt.write("{VAL},\n".format(VAL=''.join(["'''", fld_val, " '''"])))
            else:
                raise Exception("QUOTE PROBLEM PMID({}): {}".format(pmid, fld_val))
        else:
            #### prt.write("{VAL},\n".format(VAL=addquot(fld_val)))
            prt.write("{VAL},\n".format(VAL=''.join(["'", fld_val, "'"])))

    date_patterns = [
        (r'(\d{4} \S{3} \d{1,2})\s*-', '%Y %b %d'),
        (r'(\d{4} \S{3})\w?\s*-', "%Y %b"),
        (r'(\d{4} \d{2})\s*-', "%Y %m"),
        (r'(\d{4} \w{3}) - \w{3}\b', '%Y %b'),
    ]

    # pylint: disable=too-many-statements
    def _init_date(self, fld2objs, fld, str_date, pmid):
        """Convert string date to datetime object."""
        # Compensate for fmts: 1993-1994, 2001 Jul 16-31, 2002 Sep 1-15, 2012 Sep-Oct
        # TBD: Use Python Template instead of multiple replace statements?
        # https://docs.python.org/3/library/string.html#template-strings
        # pylint: disable=bad-whitespace, too-many-branches
        str_date = str_date.replace("Autumn", "Oct")
        str_date = str_date.replace("Fall",   "Oct")
        str_date = str_date.replace("Winter", "Jan")
        str_date = str_date.replace("Spring", "Apr")
        str_date = str_date.replace("Summer", "Jun")
        str_date = str_date.replace("1st Quart", "Jan")
        str_date = str_date.replace("Apr.", "Apr")
        str_date = str_date.replace("November", "Nov")
        str_date = str_date.replace("December", "Dec")
        str_date = str_date.replace("/", "-")
        if "-" in str_date:
            # "2014 Jan-Feb" -> "2014 Feb"
            if len(str_date) == 12:
                str_date = str_date[0:8]
            else:
                # "2002 Sep 1-15" -> "2002 Sep 15"
                #### mtch = match(r'(\d{4} \w{3}) \d{1,2}-(\d{1,2})', str_date)
                mtch = match(r'(\d{4} \w{3}) \d{1,2}-', str_date)
                if mtch:
                    #### str_date = ' '.join([mtch.group(1), mtch.group(2)])
                    str_date = mtch.group(1)
                else:
                    # "1993-1994" -> "1994"
                    mtch = match(r'\d{4}-(\d{4})', str_date)
                    if mtch:
                        str_date = mtch.group(1)
                    else:
                        # 1983 Jul 28-Aug 3   ->   1983 Aug 3
                        #print "WAS", str_date
                        dateobj = self._matched(self.date_patterns, str_date)
                        ## if str_date == '1984 May 31-Jun 6':
                        ##     print('SSSSSSSSSSSSSSSSSSSSSSSSSSS({})'.format(str_date), dateobj)
                        if dateobj:
                            fld2objs[fld] = dateobj
                        else:
                            raise Exception("UNRECOGNIZED FORMAT({})".format(str_date))

                        #### mtch = match(r'(\d{4} \S{3} \d{1,2})\s*-', str_date)
                        #### if mtch:
                        ####     fld2objs[fld] = datetime.datetime.strptime(mtch.group(1), "%Y %b %d")
                        #### mtch = match(r'(\d{4} \S{3})\w?\s*-', str_date)
                        #### if mtch:
                        ####     fld2objs[fld] = datetime.datetime.strptime(mtch.group(1), "%Y %b")
                        #### # 2016 09-10  ->  2016 10
                        #### mtch = match(r'(\d{4} \d{2})\s*-', str_date)
                        #### if mtch:
                        ####     fld2objs[fld] = datetime.datetime.strptime(mtch.group(1), "%Y %m")
                        #### raise Exception("UNRECOGNIZED FORMAT({})".format(str_date))
        elif "/" in str_date:
            mtch = match(r'(\d{4} \w{3})/', str_date)
            if mtch:
                fld2objs[fld] = datetime.datetime.strptime(mtch.group(1), "%Y %b")
            raise Exception("UNRECOGNIZED FORMAT({})".format(str_date))
        else:
            # Apr 2017
            mtch = match(r'(\w{3} \d{4})', str_date)
            if mtch:
                fld2objs[fld] = datetime.datetime.strptime(mtch.group(1), "%b %Y")
        #print locale.getlocale()
        # fld2objs[fld] = datetime object
        date_str = self._lendate2fmt.get(len(str_date), None)
        if date_str is None:
            ## print("STRING DATE({})".format(str_date))
            mtch = search(r'(\d{4}) \d+th (\S+)', str_date) # "2016 20th Oct"
            if mtch:
                str_date = "{YEAR} {Mon}".format(YEAR=mtch.group(1), Mon=mtch.group(2))
                date_str = self._lendate2fmt.get(len(str_date), None)
        if date_str is None:
            raise Exception("BAD FORMAT ({})".format(str_date))
        fld2objs[fld] = datetime.datetime.strptime(str_date, date_str)

    @staticmethod
    def _matched(patterns, date_str):
        """Return True if any of the date-string matched any patterns"""
        for pattern, datefmt in patterns:
            mtch = match(pattern, date_str)
            if mtch:
                return datetime.datetime.strptime(mtch.group(1), datefmt)
        return None

# pylint: disable=too-few-public-methods
class ProcessLines:
    """Pre-process PubMed text lines, get a list of tuples of fields (TI) and values"""

    mhdangle = set(['herapy', 'on & control', 'ery', 'gy', 'peutic use'])

    def __init__(self, flds):
        self.flds = flds
        self.pmid = None
        self.fld = None
        self.fldvals = None

    def process_rawlines(self, ifstrm):
        """Read raw pubmed file, put all content that extends over one line into one line."""
        pmid2fldvals = {}
        self._reset_markers()
        for line in ifstrm:
            line = line.rstrip()
            if line: # If line is not empty
                self._extract_fldvals(line)
            elif self.pmid is not None: # blank line
                pmid2fldvals[self.pmid] = self.fldvals
                self._reset_markers()
        if self.pmid is not None: # blank line
            pmid2fldvals[self.pmid] = self.fldvals # Add last pmid to results
        return pmid2fldvals

    def _reset_markers(self):
        """Reset markers back to None"""
        self.pmid = None
        self.fld = None
        self.fldvals = None

    def _extract_fldvals(self, line):
        """Read and store ONE PMID record in fldvals"""
        # Start to read a new PMID record
        # Ex: PMID- 30101408
        if line[:6] == "PMID- ":
            self.pmid = int(line[6:])
            self.fldvals = [('PMID', [line[6:]])]
        # Continue reading a PMID record
        else:
            # Beginning of a field in the record
            line_body = line[6:]
            if line[4:5] == "-":
                self.fld = line[:4].rstrip()
                if self.fld in self.flds:
                    self.fldvals.append((self.fld, [line_body]))
            # Continuation of a fields in the record
            elif self.pmid is not None and self.fld is not None and self.fld in self.flds:
                # MH is not one text block. It is a list. Update the last item on the list. Example.
                #     MH  - Single-Chain Antibodies/administration &
                #           dosage/chemistry/*immunology/pharmacokinetics
                if self.fld == 'MH':
                    sep = " " if line_body not in self.mhdangle else ""
                    # self.fldvals indexes:
                    #     [-1] -> Last MH to be added
                    #      [0] -> 'MH'; [1] -> MH string
                    #     [-1] -> Last string in the last MH field
                    self.fldvals[-1][1][-1] = sep.join([self.fldvals[-1][1][-1], line_body])
                # Text is one block of text, like the Abstracts
                else:
                    self.fldvals[-1][1].append(line_body)


  # Copyright (C) 2019-present, DV Klopfenstein. All rights reserved.



"""Report the date each TOP line was added in an iCite report file"""

__copyright__ = "Copyright (C) 2020-present, DV Klopfenstein. All rights reserved."
__author__ = "DV Klopfenstein"

import os
import sys
import collections as cx
import re
from datetime import datetime
from subprocess import Popen
from subprocess import PIPE
from argparse import ArgumentParser


# pylint: disable=too-few-public-methods
class RptDatesTop:
    """Report the date each TOP line was added in an iCite report file"""
    # Dates are retrieved using: git blame -L 1,+1 lit.txt

    datecmp = re.compile(r' (\d{4})-(\d{2})-(\d{2}) (\d{2}):(\d{2}):(\d{2})')
    weekdays = ['M', 'T', 'W', 'X', 'F', 'S', 'U']

    def __init__(self):
        self.argparser = self._init_argparser()

    @staticmethod
    def _init_argparser():
        """Add args to link pubs and iCite/PubMed"""
        parser = ArgumentParser(
            description='Report the date each paper was appended to an iCite report')
        parser.add_argument(
            'iCite_files', type=str, nargs='+',
            help='an integer for the accumulator')
        return parser

    def cli(self):
        """Run iCite/PubMed/pubs"""
        # Arguments
        args = self.argparser.parse_args()
        ## print('ARGS iCite FILE DATE-REPORTER:', args.iCite_files)
        if args.iCite_files:
            nts = self._get_line_numbers_all(args.iCite_files)
            self._prt_nts(nts)
        else:
            print('**NOTE: Please provide a file created using icite')

    @staticmethod
    def _prt_nts(nts, prt=sys.stdout):
        """Print the date when each paper was added to the iCite file"""
        pat = '{D} {DATE} {LINE}'
        for ntd in nts:
            txt = pat.format(
                D=ntd.day,
                # https://docs.python.org/2/library/datetime.html
                DATE=ntd.dateobj.strftime('%Y-%m-%d'),
                LINE=ntd.line)
            prt.write(txt)

    def _get_line_numbers_all(self, icite_files, prt=sys.stdout):
        """Get the filename and line numbers for all iCite files"""
        nts_all = []
        for fin_icite in icite_files:
            nts_file = self._get_line_numbers_one(fin_icite, prt)
            if nts_file:
                nts_all += nts_file
        return nts_all

    def _get_line_numbers_one(self, icite_file, prt=sys.stdout):
        """Get the filename and line numbers for one iCite file"""
        if not os.path.exists(icite_file):
            return []
        nts = []
        nto = cx.namedtuple('Nt', 'day dateobj lnum line filename')
        with open(icite_file) as ifstrm:
            s_weekdays = self.weekdays
            s_get_date = self._get_date
            for lnum, line in enumerate(ifstrm, 1):
                if line[:4] == 'TOP ':
                    dateobj = s_get_date(lnum, icite_file)
                    if dateobj is not None:
                        ntd = nto(
                            day=s_weekdays[dateobj.weekday()],
                            dateobj=dateobj,
                            lnum=lnum,
                            line=line,
                            filename=icite_file)
                        print(ntd)
                        nts.append(ntd)
            if prt:
                prt.write('{N:6} TOP papers READ: {FIN}'.format(N=len(nts), FIN=icite_file))
        return nts

    def _get_date(self, lnum, filename):
        """Get the date the line was changed"""
        popenargs = ['git', 'blame', '-L', '{LNUM},+1'.format(LNUM=lnum), filename]
        ## print(' '.join(popenargs))
        blame, _ = Popen(popenargs, stdout=PIPE).communicate() # _ err
        blame = blame.decode("utf-8")
        mtch = self.datecmp.search(blame)
        if mtch:
            numbers = [int(n) for n in mtch.groups()]
            return datetime(*numbers)
            ## print(blame, self.weekdays[timestamp.weekday()], timestamp)
        return None


# Copyright (C) 2020-present DV Klopfenstein. All rights reserved.



"""Given a PubMed ID (PMID), download a list of publications which cite and reference it"""
# https://icite.od.nih.gov/api

__copyright__ = "Copyright (C) 2019-present, DV Klopfenstein. All rights reserved."
__author__ = "DV Klopfenstein"

from sys import stdout
from os.path import exists

from collections import namedtuple
from collections import OrderedDict

from pmidcite.cli.utils import read_top_pmids
from pmidcite.icite.entry import NIHiCiteEntry
from pmidcite.icite.api import NIHiCiteAPI
from pmidcite.icite.nih_grouper import NihGrouper
from pmidcite.icite.paper import NIHiCitePaper


class NIHiCiteDownloaderBase:
    """Given a PubMed ID (PMID), download a list of publications which cite and reference it"""

    def __init__(self, details_cites_refs=None, nih_grouper=None):
        self.api = NIHiCiteAPI()
        # Default:set()  Options:{'cited_by_clin', 'cited_by', 'references'}
        self.details_cites_refs = self._init_details_cites_refs(details_cites_refs)
        self.nihgrouper = nih_grouper if nih_grouper is not None else NihGrouper()

    def get_icites(self, pmids):
        """Citation data should be downloaded or loaded by derived classes"""
        # pylint: disable=unreachable,no-self-use,useless-return
        raise RuntimeError("**FATAL NIHiCiteDownloaderBase:get_icites(pmids)")
        return []

    def get_icite(self, pmid):
        """Citation data should be downloaded or loaded by derived classes"""
        # pylint: disable=unreachable,no-self-use
        raise RuntimeError("**FATAL NIHiCiteDownloaderBase:get_icite(pmid)")
        return False

    def wr_papers(self, fout_txt, pmid2icitepaper, force_overwrite=False, mode='w'):
        """Run iCite for user-provided PMIDs and write to a file"""
        if not pmid2icitepaper:
            return
        pmids_all = pmid2icitepaper.keys()
        pmids_new = pmids_all
        if mode == 'a':
            pmids_new = self._get_new_pmids(fout_txt, pmids_all)
        if pmids_new:
            if self._do_write(fout_txt, force_overwrite):
                ## print('STARTING ', mode)
                with open(fout_txt, mode) as prt:
                    self.prt_papers(pmid2icitepaper, prt)
                print('{WR}: {TXT}'.format(
                    WR=self._msg_wrote(mode, pmids_all, pmids_new), TXT=fout_txt))

    def prt_api_msgs(self):
        """Print error messages captured by API, if there are any"""
        if self.api.msgs:
            print('\n{MSG}'.format(MSG='\n'.join(self.api.msgs)))

    @staticmethod
    def prt_top(paper, prt=stdout):
        """Print one detailed line summarizing the paper"""
        prt.write('TOP {iCite}\n'.format(iCite=paper.str_line()))

    def prt_papers(self, pmid2icitepaper, prt=stdout):
        """Print papers, including citation counts, cite_by and references list"""
        for pmid, paper in pmid2icitepaper.items():
            if paper is not None:
                self.prt_paper(paper, pmid, pmid, prt)
            else:
                print('**WARNING: NO iCite ENTRY FOUND FOR: {PMID}'.format(PMID=pmid))

    def prt_paper(self, paper, pmid, name, prt=stdout):
        """Print one paper, including citation counts, cite_by and references list"""
        if paper is not None:
            if self.details_cites_refs:
                paper.prt_summary(prt, sortby_cites='nih_group', sortby_refs='nih_group')
                prt.write('\n')
            else:
                self.prt_top(paper, prt)
        else:
            prt.write('No iCite results found: {PMID} {NAME}\n\n'.format(
                PMID=pmid, NAME=name if name is not None else ''))

    @staticmethod
    def _msg_wrote(mode, pmids_req, pmids_new):
        """Get the 'WROTE' or 'APPENDED' message"""
        if mode == 'w':
            return '{N:6,} WROTE'.format(N=len(pmids_req))
        if mode == 'a':
            if pmids_new:
                return '{N:,} of {M:,} APPENDED'.format(
                    N=len(pmids_new),
                    M=len(pmids_req))
            return '{N:,} of {M:,} FOUND'.format(
                N=len(pmids_new),
                M=len(pmids_req))
            # return '{N:,} FOUND'.format(M=len(pmids_req))
        raise RuntimeError('UNRECOGNIZED WRITE MODE({M})'.format(M=mode))

    @staticmethod
    def _get_new_pmids(pmidcite_txt, pmids):
        """Get PMIDs which are not already fully analyzed in pmidcite.txt"""
        if not exists(pmidcite_txt):
            return pmids
        pmids_old = read_top_pmids(pmidcite_txt)
        return [p for p in pmids if p not in pmids_old]

    def wr_name2pmid(self, fout_txt, name2pmid):
        """Run iCite for user-provided PMIDs and write to a file"""
        name2ntpaper = self._run_icite_name2pmid(name2pmid, pmid2note=None)
        if name2ntpaper:
            with open(fout_txt, 'w') as prt:
                for name, ntpaper in name2ntpaper.items():
                    self.prt_paper(ntpaper.paper, ntpaper.pmid, name, prt)
                print('  WROTE: {TXT}'.format(TXT=fout_txt))

    def _run_icite_name2pmid(self, name2pmid, pmid2note):
        """Get a NIHiCitePaper object for each user-specified PMID"""
        # TBD: Optimize for speed like _geticitepapers_w_assc
        name2ntpaper = {}
        ntobj = namedtuple('Paper', 'pmid paper')
        s_geticitepaper = self._geticitepaper
        for name, pmid in name2pmid.items():
            paper = s_geticitepaper(pmid, name, pmid2note)
            name2ntpaper[name] = ntobj(pmid=pmid, paper=paper)
        return name2ntpaper

    def get_pmid2paper(self, pmids_top, pmid2note=None):
        """Get one NIHiCitePaper object for each user-specified PMID"""
        if not self.details_cites_refs:
            return self._geticitepapers_wo_assc(pmids_top, pmid2note)
        #print('PPPPPPPPPPPPPPPPPPP', pmids_top)
        #print('PPPPPPPPPPPPPPPPPPP', pmid2note)
        return self._geticitepapers_w_assc(pmids_top, pmid2note)
        #### s_geticitepaper = self._geticitepaper
        #### papers = [s_geticitepaper(p, '', pmid2note) for p in pmids_top]
        #### return OrderedDict(zip(pmids_top, papers))  # pmid2ntpaper

    def _geticitepapers_wo_assc(self, pmids_top, pmid2note=None):
        """Get one NIHiCitePaper object for each user-specified PMID only"""
        pmid_paper = []
        pmid2icite = {o.pmid:o for o in self.get_icites(pmids_top)}
        for pmid in pmids_top:
            if pmid in pmid2icite:
                nihicite = pmid2icite[pmid]
                paper = NIHiCitePaper(pmid, {pmid:nihicite}, '', pmid2note)
                pmid_paper.append((pmid, paper))
        return OrderedDict(pmid_paper)

    def _geticitepapers_w_assc(self, pmids_top, pmid2note=None):
        """Get one NIHiCitePaper object for each user-specified PMID w/cites and/or refs"""
        nihentries_top = self.get_icites(pmids_top)
        s_get_cites_refs = self.details_cites_refs
        #print('AAAAAAAAAAAAAAAAAAAA', nihentries_top)
        top_n_assocs = [(e.pmid, e.get_assc_pmids(s_get_cites_refs)) for e in nihentries_top]
        # PMIDs that are associated with the top PMIDs (cited_by_clin, cited_by, references)
        #print('AAAAAAAAAAAAAAAAAAAA', top_n_assocs)
        #if not top_n_assocs:
        #    return {}
        pmids_assoc = set.union(*list(zip(*top_n_assocs))[1])
        # Get associated PMIDs that were requested by the researcher
        nihentries_other = self.get_icites(pmids_assoc.difference(pmids_top))
        pmid2entry_all = {e.pmid:e for e in nihentries_top + nihentries_other}
        # Create and return pmid2paper
        pmid_n_paper = []
        pmids_w_entry = set(pmid2entry_all.keys())
        header = ''
        for pmid_top, pmids_assoc in top_n_assocs:
            # pylint: disable=line-too-long
            if pmid_top in pmid2entry_all:
                pmid2entry_cur = {p:pmid2entry_all[p] for p in pmids_assoc.intersection(pmids_w_entry)}
                pmid2entry_cur[pmid_top] = pmid2entry_all[pmid_top]
                pmid_n_paper.append((pmid_top, NIHiCitePaper(pmid_top, pmid2entry_cur, header, pmid2note)))
            # Note: if there is no iCite entry for a PMID, paper will be None
            else:
                pmid_n_paper.append((pmid_top, None))
        return OrderedDict(pmid_n_paper)

    def get_paper(self, pmid, pmid2note=None):
        """Get one NIHiCitePaper object for each user-specified PMID"""
        return self._geticitepaper(pmid, header='', pmid2note=None if not pmid2note else pmid2note)

    def _geticitepaper(self, pmid_top, header, pmid2note):
        """Print summary for each user-specified PMID"""
        # pylint: disable=no-member
        top_nih_icite_entry = self.get_icite(pmid_top)  # get NIHiCiteEntry
        if top_nih_icite_entry:
            pmid2icite = {top_nih_icite_entry.pmid:top_nih_icite_entry}
            if self.details_cites_refs:
                assoc_pmids = top_nih_icite_entry.get_assc_pmids(self.details_cites_refs)
                for nihentry in self.get_icites(assoc_pmids):
                    pmid2icite[nihentry.pmid] = nihentry
            return NIHiCitePaper(pmid_top, pmid2icite, header, pmid2note)
        note = ''
        if pmid2note and pmid_top in pmid2note:
            note = pmid2note[pmid_top]
        print('No NIH iCite results found: {PMID} {HDR} {NOTE}'.format(
            PMID=pmid_top,
            HDR=header if header else '',
            NOTE=note))
        return None  ## TBD: NIHiCitePaper(pmid_top, self.dir_dnld, header, note)

    @staticmethod
    def _do_write(fout_txt, force_overwrite):
        """Ask for a yes-no answer from the user on STDIN"""
        if not exists(fout_txt) or force_overwrite:
            return True
        prompt_user = '\nover-write {TXT} (yes/no)? '.format(TXT=fout_txt)
        return input(prompt_user).lower()[:1] == 'y'

    @staticmethod
    def _init_details_cites_refs(details_cites_refs):
        """Initialize associated PMID keys"""
        # Default: Only load requested PMID (not citations or references)
        if details_cites_refs is None:
            return set()
        if details_cites_refs == 'all':
            return NIHiCiteEntry.associated_pmid_keys
        if details_cites_refs == 'citations':
            return NIHiCiteEntry.citekeys
        if details_cites_refs == 'references':
            return NIHiCiteEntry.refkey
        if isinstance(details_cites_refs, str):
            msg = ('**FATAL VALUE({V}) IN NIHiCiteDownloader(details_cites_refs="{V}"): '
                   'EXPECTED ONE OF: all citations references')
            raise RuntimeError(msg.format(V=details_cites_refs))
        return set(details_cites_refs).intersection(NIHiCiteEntry.associated_pmid_keys)


# Copyright (C) 2019-present DV Klopfenstein. All rights reserved.



"""Get a Dowloader/Loader or Downloader-Only"""

__copyright__ = "Copyright (C) 2021-present, DV Klopfenstein. All rights reserved."
__author__ = "DV Klopfenstein"

from sys import stdout
from pmidcite.icite.pmid_dnlder import NIHiCiteDownloader
from pmidcite.icite.dnldr.pmid_dnlder_only import NIHiCiteDownloaderOnly


def get_downloader(
        details_cites_refs=None,
        nih_grouper=None,
        dir_icite_py=None,
        force_download=True):
    """Get a Dowloader/Loader or Downloader-Only"""
    if not dir_icite_py or dir_icite_py == 'None':
        return NIHiCiteDownloaderOnly(details_cites_refs, nih_grouper)
    return NIHiCiteDownloader(
        dir_icite_py,
        force_download,
        details_cites_refs,
        nih_grouper)

def prt_hdr(prt=stdout):
    """Print column headers in one line"""
    from pmidcite.icite.entry import NIHiCiteEntry
    prt.write('COL {HDR}\n'.format(HDR=NIHiCiteEntry.col_idx))
    prt.write('TYP {HDR}\n'.format(HDR=NIHiCiteEntry.hdr))

def prt_keys(prt=stdout):
    """Print paper keys"""
    from pmidcite.icite.entry import NIHiCiteEntry
    from pmidcite.icite.paper import NIHiCitePaper
    prt.write('\nKEYS TO PAPER LINE:\n')
    prt.write('    TYP {ICITE_FMT}\n'.format(ICITE_FMT=NIHiCiteEntry.line_fmt()))
    prt.write('\n')
    prt.write('TYPe of relationship to the user-requested paper (TYP):\n')
    NIHiCitePaper.prt_keys(prt)
    prt.write('\nNIH iCite details:\n\n')
    NIHiCiteEntry.prt_key_desc(prt)
    prt.write('\n')


# Copyright (C) 2021-present DV Klopfenstein. All rights reserved.



#!/usr/bin/env python3
"""Given a user query, query PubMed and return PMIDs. Then run NIH's iCite on the PMIDs"""

__copyright__ = "Copyright (C) 2020-present, DV Klopfenstein. All rights reserved."
__author__ = "DV Klopfenstein"

import sys
from pmidcite.pubmedqueryicite import PubMedQueryToICite


def main():
    """Download PMIDs returned for a PubMed query. Write an iCite report for each PMID"""
    # pylint: disable=bad-whitespace
    queries = [
        # Output filenames               PubMed query
        # -----------------              -----------------------------------
        ('systematic_review.txt',        'systematic review AND "how to"[TI]'),
        ('rarely_seen_killer_whale.txt', 'Orcinus Orca Type D'),
    ]

    # By default, only the last entry in the list is run.
    # This allows you to build a history of searches,
    # but not run all of them every time.
    #
    # To re-run all entries in the list:
    #   $ src/bin/dnld_pmids.py all
    #
    # To run the first query:
    #   $ src/bin/dnld_pmids.py 0
    #
    # To run the second to last query:
    #   $ src/bin/dnld_pmids.py -2
    #
    obj = PubMedQueryToICite(force_dnld=True)
    dnld_idx = obj.get_index(sys.argv)
    obj.run(queries, dnld_idx)


if __name__ == '__main__':
    main()

# Copyright (C) 2020-present DV Klopfenstein. All rights reserved.



#!/usr/bin/env python3
"""Given a PubMed ID (PMID), download its record from PubMed"""

__copyright__ = "Copyright (C) 2019-present, DV Klopfenstein. All rights reserved."
__author__ = "DV Klopfenstein"

from pmidcite.cli.dnldpubmed import DnldPubMed  # get_argparser


def main():
    """Given a PubMed ID (PMID), download its record from PubMed"""
    # PubMed records include the abstract, list of authors, journal, etc.
    DnldPubMed().cli()


if __name__ == '__main__':
    main()

# Copyright (C) 2019-present, DV Klopfenstein. All rights reserved.



#!/usr/bin/env python3
"""ASCII scatter plot adapted from https://github.com/dzerbino/ascii_plots"""

__copyright__ = "Copyright (C) 2013-present, Authored by Daniel Zerbino. All rights reserved."
__copyright__ = "Copyright (C) 2020-present, Adapted for pmidcite by DV Klopfenstein. All rights reserved."


import sys
from pmidcite.plot.scatter import AsciiScatter


def main():
    """ASCII scatter plot adapted from https://github.com/dzerbino/ascii_plots"""
    obj = AsciiScatter()
    xydata = obj.read_stdin_ints()
    obj.run(xydata)


if __name__ == '__main__':
    main()

# Copyright (C) 2020-present DV Klopfenstein. All rights reserved.



"""Run PubMed user query and download PMIDs. Run iCite on PMIDs. Write text file."""

__copyright__ = "Copyright (C) 2019-present, DV Klopfenstein. All rights reserved."
__author__ = "DV Klopfenstein"

import sys
from collections import namedtuple
from pmidcite.cfg import get_cfgparser
from pmidcite.eutils.cmds.pubmed import PubMed
#### from pmidcite.cli.utils import wr_pmids
from pmidcite.icite.downloader import get_downloader


class PubMedQueryToICite:
    """Run PubMed user query and download PMIDs. Run iCite on PMIDs. Write text file."""

    def __init__(self, force_dnld, verbose=True, pmid2note=None):
        self.force_dnld = force_dnld
        self.verbose = verbose
        self.pmid2note = {} if pmid2note is None else pmid2note
        self.cfg = get_cfgparser()
        self.pubmed = PubMed(
            email=self.cfg.get_email(),
            apikey=self.cfg.get_apikey(),
            tool=self.cfg.get_tool())

    def run(self, fout_query, dnld_idxs=None):
        """Give an list of tuples: Query PubMed for PMIDs. Download iCite, given PMIDs"""
        # Download PMIDs and NIH's iCite for only one PubMed query
        nts = self.get_nts_g_list(fout_query)
        if dnld_idxs is not None:
            for idx in dnld_idxs:
                ntd = nts[idx]
                self.querypubmed_runicite(ntd.filename, ntd.pubmed_query)
        # Download PMIDs and NIH's iCite for all PubMed queries
        else:
            for ntd in nts:
                self.querypubmed_runicite(ntd.filename, ntd.pubmed_query)
        return nts

    def run_one(self, fout_query, dnld_idx):
        """Query PubMed for PMIDs. Download iCite, given PMIDs"""
        nts = self.get_nts_g_list(fout_query)
        ntd = nts[dnld_idx]
        return self.querypubmed_runicite(ntd.filename, ntd.pubmed_query)

    def querypubmed_runicite(self, filename, query, details_cites_refs=None):
        """Given a user query, return PMIDs. Then run NIH's iCite"""
        # 1) Query PubMed and download PMIDs
        #    Maximum PMIDs to download at a time is 100,000 (Default is 20):
        #        PMIDCITE uses usehistory='y', so can retrieve more than 100,000 PMIDs
        #        by incrementing the value of retstart for each 100k set of PMIDs
        #        https://www.ncbi.nlm.nih.gov/books/NBK25499/#chapter4.ESearch
        pmids = self.pubmed.dnld_query_pmids(query, num_ids_p_epost=100000)
        ## print('PMIDCITE QQQQQQQQQQQQQQQQQ {N} PMIDs'.format(N=len(pmids)))
        ## fout_pmids = self.cfg.get_fullname_pmids(filename)
        fout_icite = self.cfg.get_fullname_icite(filename)
        ##### 2) Write PubMed PMIDs into a simple text file, one PMID per line
        ####if fout_pmids != fout_icite:
        ####    if pmids:
        ####        wr_pmids(fout_pmids, pmids)
        ####    else:
        ####        print('  0 PMIDs: NOT WRITING {TXT}'.format(TXT=fout_pmids))
        # 3) Run NIH's iCite on the PMIDs and write the results into a file
        if pmids:
            return self._wr_icite(fout_icite, pmids, details_cites_refs)
        return {}

    def _wr_icite(self, fout_icite, pmids, details_cites_refs):
        """Run PMIDs in iCite and print results into a file"""
        cfg = self.cfg
        dnldr = get_downloader(
            # all citations references
            details_cites_refs=details_cites_refs,
            nih_grouper=cfg.get_nihgrouper(),
            dir_icite_py=cfg.get_dir_icite_py(),
            force_download=self.force_dnld)
        ## print('PMIDCITE PPPPPPPPPPPPPPPPP dnldr.get_pmid2paper {N} PMIDs'.format(N=len(pmids)))
        pmid2paper = dnldr.get_pmid2paper(pmids, self.pmid2note)
        ## print('PMIDCITE PPPPPPPPPPPPPPPPP dnldr.wr_papers{N} PMIDs'.format(N=len(pmids)))
        dnldr.wr_papers(fout_icite, pmid2icitepaper=pmid2paper, force_overwrite=True)
        return pmid2paper

    def get_nts_g_list(self, lst):
        """Turn a list iof tuple strings into a list of namedtuples"""
        nto = namedtuple('Nt', 'filename fullname pubmed_query')
        return [nto._make([fname, self.cfg.get_fullname_icite(fname), qry]) for fname, qry in lst]

    @staticmethod
    def get_index(argv, queries=None):
        """Get the index of the pubmed query to run"""
        # If no argument was provided, run the last query in the list
        if len(argv) == 1:
            return [-1]
        if 'h' in argv[1] and queries:
            for idx, item in enumerate(queries):
                print('{I:3} {E}'.format(I=idx, E=item))
            sys.exit(1)
        return [int(n) for n in argv[1:] if n.lstrip('-').isdigit()]


# Copyright (C) 2019-present DV Klopfenstein. All rights reserved.



"""Holds NIH iCite data for one PubMed ID (PMID)"""
# https://icite.od.nih.gov/api

__copyright__ = "Copyright (C) 2019-present, DV Klopfenstein. All rights reserved."
__author__ = "DV Klopfenstein"

from sys import stdout


class NIHiCiteEntry:
    """Holds NIH iCite data for one PubMed ID (PMID)"""

    fields = ['pmid', 'aart_type', 'aart_animal', 'nih_perc', 'nih_group', 'year',
              'num_cites_all', 'clin', 'references',
              'A', 'author1', 'title']

    pat_pre = ('{pmid:8} {aart_type:2} {aart_animal:5} '
               '{nih_perc:>3} {nih_group} {year:4} {num_cites_all:>5} {clin:>2} {references:>3} '
              )
    pat_str = pat_pre + 'au[{A:02}]({author1}) {title}'

    pat_md = ('{year}|{pmid:8}|{aart_type}|{aart_animal}|'
              '{num_cites_all:>5}|{clin:>4}|{references:>4}|'
              '{A:2}|{author1:16}|{title}'
             )

    citekeys = {'cited_by_clin', 'cited_by'}
    refkey = {'references',}
    associated_pmid_keys = citekeys.union(refkey)

    col_idx = pat_str.format(
        pmid='2',
        aart_type='3',
        aart_animal='4',
        nih_perc='5',
        nih_group='6',
        year='7',
        num_cites_all='8',
        clin='9',
        references='10',
        A=11,
        author1='authors',
        title='')

    hdr = pat_str.format(
        pmid='PMID',
        year='YEAR',
        aart_type='RP',
        aart_animal='HAMCc',
        nih_group='G',
        nih_perc='  %',
        num_cites_all='  cit',
        clin='cli',
        references='ref',
        A=0,
        author1='authors',
        title='title')

    def __init__(self, icite_dct, nih_group):
        self.pmid = icite_dct['pmid']
        self.dct = icite_dct
        nih_perc = icite_dct['nih_percentile']
        self.dct['nih_group'] = nih_group  # 0 - 5
        # pylint: disable=line-too-long
        self.dct['num_auth'] = len(icite_dct['authors'])
        self.dct['num_clin'] = len(icite_dct['cited_by_clin'])
        self.dct['num_cite'] = len(icite_dct['cited_by'])
        num_cites_all = len(set(self.dct['cited_by_clin']).union(self.dct['cited_by']))
        self.dct['num_cites_all'] = num_cites_all
        self.dct['nih_perc'] = round(nih_perc) if nih_perc is not None else 110 + num_cites_all
        self.dct['num_refs'] = len(icite_dct['references'])

    ## TBD:
    ## def __eq__(self, rhs):
    ##     if self.pmid == rhs.pmid:
    ##         return False
    ##     return True

    def get_au1_lastname(self):
        """Get the last name of the first author"""
        aus = self.dct['authors']
        if aus:
            flds = aus[0].split()
            return flds[-1]
        return None

    def get_year(self):
        """Get the publication year"""
        return self.dct.get('year')

    def prt_cite_cnts(self):
        """Print cite counts. NIH's citation_count includes only cited_by, not cited_by_clin"""
        # pylint: disable=line-too-long
        # if self.dct['citation_count'] != len(set(self.dct['cited_by'])):
        # if self.dct['citation_count'] != len(set(self.dct['cited_by_clin']).union(self.dct['cited_by']))
        cited_by_clin = set(self.dct['cited_by_clin'])
        cited_by = set(self.dct['cited_by'])
        num_exp = len(cited_by_clin.union(cited_by))
        msg = ('**NOTE: {pmid:9}=pmid '
               '{citation_count:6,}=citation_count '
               '{num_clin:6,}=cited_by_clin '
               '{num_cite:6,}=cited_by;   '
               '{num_exp:6,}=cited_by_clin.union(cited_by)')
        print(msg.format(num_exp=num_exp, **self.dct))

    def prt_keys(self, prt=stdout):
        """Print paper keys, including header line"""
        prt.write('{COLS}\n\n'.format(COLS=self.line_fmt()))
        self.prt_key_desc(prt)

    @staticmethod
    def line_fmt():
        """Return the format of the paper line"""
        return 'PubMedID RP HAMCc % G YEAR x y z au[A](First Author) Title of paper'

    @staticmethod
    def prt_key_desc(prt=stdout):
        """Print paper key description"""
        # pylint: disable=line-too-long
        ## prt.write('NIH iCite line format:\n')
        ## prt.write('  YYYY NNNNNNNN RP HAMCc nih% x y z au[A](First Author) Title of paper\n\n')
        ## prt.write('NIH iCite details:\n')
        prt.write('  PubMedID: PubMed ID (PMID)\n\n')
        prt.write('     RP section:\n')
        prt.write('     ----------------------------------\n')
        prt.write('         R: Is a research article\n')
        prt.write('         P: iCite has calculated an initial Relative Citation Ratio (RCR) for new papers\n\n')
        prt.write('     HAMCc section:\n')
        prt.write('     ----------------------------------\n')
        prt.write('         H: Has MeSH terms in the human category\n')
        prt.write('         A: Has MeSH terms in the animal category\n')
        prt.write('         M: Has MeSH terms in the molecular/cellular biology category\n')
        prt.write('         C: Is a clinical trial, study, or guideline\n')
        prt.write('         c: Is cited by a clinical trial, study, or guideline\n\n')
        prt.write('     NIH section, based on Relative Citation Ratio (RCR):\n')
        prt.write('     ----------------------------------\n')
        prt.write('         %: NIH citation percentile rounded to an integer. -1 means "not determined" or TBD\n')
        prt.write('         G: NIH citation percentile group: 0=-3SD 1=-2SD 2=+/-1SD 3=+2SD 4=+3SD or i=TBD\n\n')
        prt.write('     YEAR/citations/references section:\n')
        prt.write('     ----------------------------------\n')
        prt.write('      YEAR: The year the article was published\n')
        prt.write('         x: Total of all unique articles that have cited the paper, including clinical articles\n')
        prt.write('         y: Number of unique clinical articles that have cited the paper\n')
        prt.write('         z: Number of references\n')
        prt.write('     au[A]: A is the number of authors\n')

    def str_md(self):
        """Return one-line string describing NIH iCite entry"""
        return self._str(self.pat_md)

    def __str__(self):
        """Return one-line string describing NIH iCite entry"""
        return self._str(self.pat_str)

    def _str(self, pat):
        """Return one-line string describing NIH iCite entry"""
        dct = self.dct
        nih_group = dct['nih_group']
        nih_perc = dct['nih_perc']
        return pat.format(
            pmid=self.pmid,
            year=dct['year'],
            aart_type=self.get_aart_type(),
            aart_animal=self.get_aart_translation(),
            nih_group=str(nih_group) if nih_group != 5 else 'i',
            nih_perc=nih_perc if nih_perc <= 100 else ' -1',
            num_cites_all=dct['num_cites_all'],
            clin=dct['num_clin'],
            references=dct['num_refs'],
            A=dct['num_auth'],
            author1=dct['authors'][0] if dct['authors'] else '',
            title=dct['title'],
        )

    def get_assc_pmids(self, keys):
        """Get PMIDs associated with the given NIH iCite data"""
        pmids = set()
        s_dct = self.dct
        for assc_key in keys:
            if s_dct[assc_key]:
                pmids.update(s_dct[assc_key])
        return pmids

    def get_aart_type(self):
        """Get succinct ASCII art for concise info display"""
        lst = []
        dct = self.dct
        lst.append('R' if dct['is_research_article'] else '.')
        lst.append('P' if dct['provisional'] else '.')
        return ''.join(lst)

    def get_aart_translation(self):
        """Get succinct ASCII art for the translation information"""
        # Translation of basic sciene research into practical clinical applications
        # or 'bench-to-bedside'
        # https://www.ncbi.nlm.nih.gov/pubmed/23705970
        # https://journals.plos.org/plosbiology/article?id=10.1371/journal.pbio.3000416
        lst = []
        dct = self.dct
        lst.append('H' if dct['human'] != 0.0 else '.')
        lst.append('A' if dct['animal'] != 0.0 else '.')
        lst.append('M' if dct['molecular_cellular'] != 0.0 else '.')
        lst.append('C' if dct['is_clinical'] else '.')
        lst.append('c' if dct['cited_by_clin'] else '.')
        return ''.join(lst)

    def prt_dct(self, prt=stdout):
        """Print full NIH iCite dictionary"""
        prt.write('{STR}\n'.format(STR=self.str_dct()))

    def str_dct(self):
        """Get a string describing object"""
        txt = []
        for key, val in self.dct.items():
            if isinstance(val, list):
                val = '[{N}] {LST}'.format(N=len(val), LST=', '.join(str(e) for e in val))
            txt.append('{K:27} {V}'.format(K=key, V=val))
        return '\n'.join(txt)

    def __lt__(self, rhs):
        """Default sort by PMID"""
        return self.pmid < rhs.pmid


# Copyright (C) 2019-present DV Klopfensteinr,. All rights reserved.



"""Manage MeSH descriptors or qualifiers stored in Python modules."""

__copyright__ = "Copyright (C) 2019-present, DV Klopfenstein. All rights reserved."
__author__ = "DV Klopfenstein"

import sys
import collections as cx
from pmidcite.eutils.pubmed.descriptors import UI2NT
from pmidcite.eutils.pubmed.qualifiers import NTS as nts_mesh_qual

class MeshTerms(object):
  """Manage MeSH descriptors or qualifiers stored in Python modules."""

  def __init__(self):
    # nt: MH MN UI;    Ex 'Abdomen', set(['A01.923.047']), 'D000005'
    self.mh2nt = {nt.MH:nt for nt in UI2NT.values()}
    # nt: SH QA QE MS; Ex 'agonists', 'AG', 'AGON', 'Used with ...
    self.sh2nt = {nt.SH:nt for nt in nts_mesh_qual}
    assert len(self.mh2nt) == len(UI2NT)
    assert len(self.sh2nt) == len(nts_mesh_qual)

  def mrg_starred_mesh(self, mhphrase2cnt): # phrase_list):
    """ *genetics & genetics -> *genetics."""
    phrase_new = cx.Counter()
    # Find all starred phrases. Remove '*' before storing.
    starred = set([w for w in mhphrase2cnt.keys() if w[:1] == '*'])
    for mesh_phrase, cnt in mhphrase2cnt.items():
      star, bare_phrase = self.get_star_phrase(mesh_phrase)
      if bare_phrase in starred:
        phrase_new["".join(["*", bare_phrase])] += cnt
      else:
        phrase_new[bare_phrase] += cnt
    return phrase_new 

  def get_star_phrase(self, phrase):
    """Return Corrected MeSH term."""
    return ['', phrase] if phrase[0]!='*' else ['*', phrase[1:]]

  def get_descqual2ctr(self, mh_vals):
    """Given MH raw strings, check strings include either MeSH descriptors or qualifiers."""
    # VAL Antineoplastic Agents, Phytogenic/administration & dosage
    # VAL Apoptosis/*drug effects
    mhtype2ctr = {'desc':cx.Counter(), 'qual':cx.Counter()}
    for mhline in mh_vals:
      for mhstr in mhline.split('/'):
        _, phrase = self.get_star_phrase(mhstr)
        is_mh = phrase in self.mh2nt
        is_sh = phrase in self.sh2nt
        if is_mh and not is_sh:
          mhtype2ctr['desc'][mhstr] += 1
        elif not is_mh and is_sh:
          mhtype2ctr['qual'][mhstr] += 1
        else:
          raise Exception("UNKNOWN VALUE({}) IN ({})".format(mhstr, mhline))
    return mhtype2ctr

  def chk_mhs(self, mh_vals):
    """Given MH raw strings, check strings include either MeSH descriptors or qualifiers."""
    # VAL Antineoplastic Agents, Phytogenic/administration & dosage
    # VAL Apoptosis/*drug effects
    for mhline in mh_vals:
      for mhstr in mhline.split('/'):
        star, phrase = self.get_star_phrase(mhstr)
        is_mh = phrase in self.mh2nt
        is_sh = phrase in self.sh2nt
        # assert is_mh != is_sh, "UNKNOWN VALUE({})".format(mhstr)
        if is_mh == is_sh:
            sys.stdout.write("MeSH CHECK-UNKNOWN VALUE({})\n".format(mhstr))


# Copyright (C) 2019-present, DV Klopfenstein. All rights reserved.



#!/usr/bin/env python3
"""Given a user query, query PubMed and return PMIDs. Then run NIH's iCite on the PMIDs"""

__copyright__ = "Copyright (C) 2020-present, DV Klopfenstein. All rights reserved."
__author__ = "DV Klopfenstein"

from pmidcite.cli.rptdatestop import RptDatesTop


def main():
    """Download PMIDs returned from user queries. Write: ./log/pmids ./log/icite"""
    obj = RptDatesTop()
    obj.cli()


if __name__ == '__main__':
    main()

# Copyright (C) 2020-present DV Klopfenstein. All rights reserved.



"""Downloaded PubMed count data"""

DATE = '2020_01_11'

# pylint: disable=line-too-long
# 30,527,422 all                  all [sb]
#  1,819,828 all_ml0_pmc0         all [sb] NOT inprocess[sb] NOT medline[sb] NOT pubmed pmc[sb]
#  1,694,611 all_ml0_pmc0_free0   all [sb] NOT inprocess[sb] NOT medline[sb] NOT pubmed pmc[sb] NOT free full text[sb]
#    125,217 all_ml0_pmc0_free1   all [sb] NOT inprocess[sb] NOT medline[sb] NOT pubmed pmc[sb] AND free full text[sb]
#  1,817,769 all_ml0_pmc0b        all [sb] NOT inprocess[sb] NOT medline[sb] NOT pubmed pmc[sb] NOT pubstatusnihms NOT pubstatuspmcsd
#  1,503,281 all_ml0_pmc0d        all [sb] NOT inprocess[sb] NOT medline[sb] NOT pubmed pmc[sb] NOT pubstatusnihms NOT pubstatuspmcsd AND pubmednotmedline[sb]
#    314,488 all_ml0_pmc0c        all [sb] NOT inprocess[sb] NOT medline[sb] NOT pubmed pmc[sb] NOT pubstatusnihms NOT pubstatuspmcsd NOT pubmednotmedline[sb]
# 28,707,594 ml1_pmc1             inprocess[sb] OR medline[sb] OR pubmed pmc[sb]
#    311,051 pub_init0            publisher[sb] NOT pubstatusnihms NOT pubstatuspmcsd NOT pmcbook
#     76,962 pub_init1            publisher[sb] AND (pubstatusnihms OR pubstatuspmcsd OR pmcbook)
#    388,013 pub                  publisher[sb]
#     31,420 nihms_pub0           pubstatusnihms AND publisher[sb]
#     42,782 nihms_pub1           pubstatusnihms NOT publisher[sb]
#     74,202 nihms                pubstatusnihms
#     72,580 nihms_pmc1           pubstatusnihms AND pubmed pmc[sb]
#      1,622 nihms_pmc0           pubstatusnihms NOT pubmed pmc[sb]
#     23,278 pub_pmcsd            pubstatuspmcsd AND publisher[sb]
#     60,153 pmcsd                pubstatuspmcsd
#          0 pmcsd_and_nihms      pubstatuspmcsd AND pubstatusnihms
#     59,274 pmcsd_pmc1           pubstatuspmcsd AND pubmed pmc[sb]
#        879 pmcsd_pmc0           pubstatuspmcsd NOT pubmed pmc[sb]
#    558,119 inprocess_A_all      inprocess[sb]
#    421,966 inprocess_A_pmc0     inprocess[sb] NOT pubmed pmc[sb]
#    136,153 inprocess_A_pmc1     inprocess[sb] AND pubmed pmc[sb]
#    558,119 inprocess_ml0        inprocess[sb] NOT medline[sb]
#          0 inprocess_ml1        inprocess[sb] AND medline[sb]
# 27,019,907 medline_n_inprocess  medline[sb] OR inprocess[sb]
# 26,461,788 medline_all          medline[sb]
# 22,956,376 medline_pmc0         medline[sb] NOT pubmed pmc[sb]
#  3,505,412 medline_pmc1         medline[sb] AND pubmed pmc[sb]
#  5,329,252 pmc_all              pubmed pmc[sb]
#  5,255,875 pmc_all?             pubmed pmc[sb] AND (inprocess[sb] OR medline[sb] OR pubmednotmedline[sb])
#  3,641,565 pmc_m1               pubmed pmc[sb] AND (inprocess[sb] OR medline[sb])
#  1,687,687 pmc_m0               pubmed pmc[sb] NOT inprocess[sb] NOT medline[sb]
#  1,614,310 pmc_m0b              pubmed pmc[sb] AND pubmednotmedline[sb]
#     73,377 pmc_unknown          pubmed pmc[sb] NOT inprocess[sb] NOT medline[sb] NOT pubmednotmedline[sb]
#    762,137 au_all               author manuscript all[sb]
#    762,137 au_all1              author manuscript all[sb] AND all[sb]
#    666,598 au_ss                author manuscript[sb] NOT pubstatusnihms NOT pubstatuspmcsd NOT pmcbook
#     34,746 au_ss_m0             author manuscript[sb] NOT pubstatusnihms NOT pubstatuspmcsd NOT pmcbook NOT medline[sb]
#      9,066 au_ss_ml0            author manuscript[sb] NOT pubstatusnihms NOT pubstatuspmcsd NOT pmcbook NOT medline[sb] NOT inprocess[sb]
#      7,106 au_ss_pmconly        author manuscript[sb] AND pubmed pmc[sb] NOT pubstatusnihms NOT pubstatuspmcsd NOT pmcbook NOT medline[sb] NOT inprocess[sb]
#     69,494 au_nihms             author manuscript[sb] AND pubstatusnihms
#         12 au_pmcsd             author manuscript[sb] AND pubstatuspmcsd
#          0 au_pmcbook           author manuscript[sb] AND pmcbook
#    659,219 au_ml1               author manuscript[sb] AND (medline[sb] OR inprocess[sb])
#     76,885 au_ml0               author manuscript[sb] NOT medline[sb] NOT inprocess[sb]
#      3,424 au_ml0_pmc0          author manuscript[sb] NOT medline[sb] NOT inprocess[sb] NOT pubmed pmc[sb]
#     73,461 au_ml0_pmc1          author manuscript[sb] AND pubmed pmc[sb] NOT medline[sb] NOT inprocess[sb]
#    633,507 au_m1                author manuscript[sb] AND medline[sb]
#     25,712 au_ip1               author manuscript[sb] AND inprocess[sb]
#    712,329 au_ml1_or_pmc1       author manuscript[sb] AND (medline[sb] OR pubmed pmc[sb])
#    625,458 au_ml1_and_pmc1      author manuscript[sb] AND medline[sb] AND pubmed pmc[sb]
#    630,819 au_ml1_and_pmc1_or_ip1 author manuscript[sb] AND (inprocess[sb] OR medline[sb] AND pubmed pmc[sb])
#    704,280 au_pmc1              author manuscript[sb] AND pubmed pmc[sb]
#      5,361 au_pmc1_ip1          author manuscript[sb] AND pubmed pmc[sb] AND inprocess[sb]
#     78,822 au_pmc1_ml0          author manuscript[sb] AND pubmed pmc[sb] NOT medline[sb]
#     31,824 au_pmc0              author manuscript[sb] NOT pubmed pmc[sb]
#      8,049 au_embargo           author manuscript[sb] AND medline[sb] NOT pubmed pmc[sb]
#  3,119,502 pmnml_A              pubmednotmedline[sb]
#  3,119,502 pmnml_B              pubmednotmedline[sb] NOT medline[sb]
#  1,505,192 pmnml_C              pubmednotmedline[sb] NOT pubmed pmc[sb]
#  1,505,192 pmnml_D              pubmednotmedline[sb] NOT pubmed pmc[sb] NOT pubstatusnihms[sb] NOT pubstatuspmcsd[sb]
#  3,119,502 pmnml_C_ip0          pubmednotmedline[sb] NOT inprocess[sb]
#          0 pmnml_0_ip1          pubmednotmedline[sb] AND inprocess[sb]
#  1,614,310 pmnml_A_pmc1         pubmednotmedline[sb] AND pubmed pmc[sb]
#  1,505,192 pmnml_A_pmc0         pubmednotmedline[sb] NOT pubmed pmc[sb]
#     22,264 pmcbook_A            pmcbook
#     22,264 pmcbook_A_ml0        pmcbook NOT medline[sb]
#     22,264 pmcbook_A_pmc0       pmcbook NOT pubmed pmc[sb]

CNTS = {
    "all": 30527422,
    "all_ml0_pmc0": 1819828,
    "all_ml0_pmc0_free0": 1694611,
    "all_ml0_pmc0_free1": 125217,
    "all_ml0_pmc0b": 1817769,
    "all_ml0_pmc0d": 1503281,
    "all_ml0_pmc0c": 314488,
    "ml1_pmc1": 28707594,
    "pub_init0": 311051,
    "pub_init1": 76962,
    "pub": 388013,
    "nihms_pub0": 31420,
    "nihms_pub1": 42782,
    "nihms": 74202,
    "nihms_pmc1": 72580,
    "nihms_pmc0": 1622,
    "pub_pmcsd": 23278,
    "pmcsd": 60153,
    "pmcsd_and_nihms": 0,
    "pmcsd_pmc1": 59274,
    "pmcsd_pmc0": 879,
    "inprocess_A_all": 558119,
    "inprocess_A_pmc0": 421966,
    "inprocess_A_pmc1": 136153,
    "inprocess_ml0": 558119,
    "inprocess_ml1": 0,
    "medline_n_inprocess": 27019907,
    "medline_all": 26461788,
    "medline_pmc0": 22956376,
    "medline_pmc1": 3505412,
    "pmc_all": 5329252,
    "pmc_all?": 5255875,
    "pmc_m1": 3641565,
    "pmc_m0": 1687687,
    "pmc_m0b": 1614310,
    "pmc_unknown": 73377,
    "au_all": 762137,
    "au_all1": 762137,
    "au_ss": 666598,
    "au_ss_m0": 34746,
    "au_ss_ml0": 9066,
    "au_ss_pmconly": 7106,
    "au_nihms": 69494,
    "au_pmcsd": 12,
    "au_pmcbook": 0,
    "au_ml1": 659219,
    "au_ml0": 76885,
    "au_ml0_pmc0": 3424,
    "au_ml0_pmc1": 73461,
    "au_m1": 633507,
    "au_ip1": 25712,
    "au_ml1_or_pmc1": 712329,
    "au_ml1_and_pmc1": 625458,
    "au_ml1_and_pmc1_or_ip1": 630819,
    "au_pmc1": 704280,
    "au_pmc1_ip1": 5361,
    "au_pmc1_ml0": 78822,
    "au_pmc0": 31824,
    "au_embargo": 8049,
    "pmnml_A": 3119502,
    "pmnml_B": 3119502,
    "pmnml_C": 1505192,
    "pmnml_D": 1505192,
    "pmnml_C_ip0": 3119502,
    "pmnml_0_ip1": 0,
    "pmnml_A_pmc1": 1614310,
    "pmnml_A_pmc0": 1505192,
    "pmcbook_A": 22264,
    "pmcbook_A_ml0": 22264,
    "pmcbook_A_pmc0": 22264,
}



"""Module helper"""

__copyright__ = "Copyright (C) 2019-present, DV Klopfenstein. All rights reserved."
__author__ = "DV Klopfenstein"

import sys
import pkgutil
import importlib
import importlib.util


def import_var(modulestr, varname, log=sys.stdout, rpterr=True):
    """Return variable inside module."""
    mod = import_mod(modulestr, log)
    if mod is not None:
        var = getattr(mod, varname, None)
        if var is not None:
            return var
    if rpterr:
        _rpt_err(mod, varname, modulestr)
    return None

def _rpt_err(mod, varname, modulestr):
    if mod is None:
        raise Exception("MODULE({M}) NOT IMPORTED".format(M=modulestr))
    raise Exception("VAR({V}) NOT FOUND IN MOD({M})".format(V=varname, M=modulestr))

def import_mod(modulestr, log=None):
    """Import Python module"""
    if log is not None:
        log.write("  IMPORT {MOD}\n".format(MOD=modulestr))
    if pkgutil.find_loader(modulestr) is not None:
        return importlib.import_module(modulestr)
    if log is not None:
        log.write("  None   {MOD}\n".format(MOD=modulestr))
    return None

def load_modpy(fin_py):
    """Load NIH iCite information from Python modules"""
    try:
        spec = importlib.util.spec_from_file_location("module.name", fin_py)
        mod = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(mod)
        return mod
    except RuntimeError as inst:
        print('**ERROR: UNABLE TO READ: {PY}\nMSG: {M}'.format(PY=fin_py, M=str(inst)))
        return None


# Copyright (C) 2019-present DV Klopfensteinr,. All rights reserved.



"""Small utilities used in the pmidcite project"""

__copyright__ = "Copyright (C) 2021-present, DV Klopfenstein. All rights reserved."
__author__ = "DV Klopfenstein"


def split_list(iterable, len_sublists):
    """Split a long list into smaller lists"""
    # WAS: [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    # NOW: [(1, 2, 3), (4, 5, 6), (7, 8, 9), (10,)]
    #
    # WAS: [1, 2, 3, 4, 5, 6, 7, 8, 9]
    # NOW: [(1, 2, 3), (4, 5, 6), (7, 8, 9)]
    lst = list(iterable)
    mod = len(iterable)%len_sublists
    ret = list(zip(*[iter(lst)]*len_sublists))
    if mod != 0:
        ret.append(tuple(lst[-1*mod:]))
    return ret


# Copyright (C) 2021-present DV Klopfenstein. All rights reserved.



#!/usr/bin/env python
"""Plot NIH percentile groups for citation data from a paper's cocitation network"""

__copyright__ = "Copyright (C) 2020-present, DV Klopfenstein. All rights reserved."


# pylint: disable=wrong-import-position
import math
import matplotlib as mpl
mpl.use('agg')
import matplotlib.pyplot as plt
import scipy.stats as stats
from pmidcite.plot.nih_perc import PltNihVals
from pmidcite.cfg import Cfg

MU = 0
VARIANCE = 1
SIGMA = math.sqrt(VARIANCE)

def main():
    """Plot guassian distribution, showing icite groups"""
    fout_imgs = [
        '../pmidcite/doc/images/nih_perc_groups.png',
        '../bibliometrics/doc/nih_perc_groups.tiff',
        '../bibliometrics/doc/nih_perc_groups.pdf',
        'nih_perc_groups.svg',
        '../bibliometrics/doc/nih_perc_groups.pdf',
    ]

    # 1a) Get percentile group_mins along SD lines
    # percentiles: [2.28, 15.87, 84.13, 97.72]
    vlines = [-3, -2, -1, 1, 2, 3]
    percentiles = [round(stats.norm.cdf(z)*100, 2) for z in vlines[1:-1]]

    # 1b) Get percentile group_mins from Cfg
    cfg = Cfg()
    nih_grouper = cfg.get_nihgrouper()
    percentiles = nih_grouper.get_list()

    # Plot Gaussian curve
    pltr = PltNihVals(percentiles)
    ## pltr = PltNihVals(Cfg().get_nihgrouper().get_list())
    pltr.prt_vals()

    fig, axes = plt.subplots(1, 1)
    pltr.plt_lines(axes)
    pltr.colorfill_groups(axes)
    pltr.add_text_groups(axes)
    pltr.anno_groupmin(axes)
    axes.set_xlabel('Standard Deviations', fontsize=15)
    axes.set_ylabel('Distribution of co-citation network', fontsize=15)

    for fout_img in fout_imgs:
        fig.tight_layout()
        plt.savefig(fout_img, dpi=800)
        print('WROTE: {IMG}'.format(IMG=fout_img))


if __name__ == '__main__':
    main()

# Copyright (C) 2020-present, DV Klopfenstein. All rights reserved.



"""Fetch items and write"""

__author__ = 'DV Klopfenstein'
__copyright__ = "Copyright (C) 2016-present DV Klopfenstein. All rights reserved."
__license__ = "GPL"

import sys
import re
from pmidcite.eutils.cmds.cmdbase import CommandBase


# TBD:
class ELink(CommandBase):
    """Fetch and write text"""

    # pylint: disable=too-many-arguments
    def __init__(self, retmax=10000, rettype='medline', retmode='text', batch_size=100, **kws):
        kws_base = {k:v for k, v in kws.items() if k in CommandBase.exp_kws}
        super(ELink, self).__init__(**kws_base)

    def elink(self, database_from, linkname, webenv, querykey, num_fetches):
        """EFetch records found for PMIDs, page by page"""
        ## QueryKey(     1) EFetching(database=pubmed) up to    10 records, starting at 0; ABSTRACT
        ## QueryKey(     1) EFetching(database=pubmed) up to    10 records, starting at 10; ABSTRACT
        ## msg_fmt = ('  QueryKey({:>6}) EFetching(database={}) up to {:5} records, '
        ##            'starting at {}; {}\n')
        for start in range(0, num_fetches, self.batch_size):
            ## msg = msg_fmt.format(querykey, database, self.batch_size, start, self.desc)
            ## sys.stdout.write(msg)
            record = None
            try:
                # pylint: disable=bad-whitespace
                record = self.run_eutilscmd(
                    'elink',
                    db        = database_from,
                    retstart  = start,       # dflt: 1
                    retmax    = self.batch_size,  # max: 10,000
                    retmode   = 'json',
                    linkname  = linkname,
                    webenv    = webenv,
                    query_key = querykey)
                print('ELINK:', linkname, record)
            except IOError as err:
                msg = "\n*FATAL: EFetching FAILED: {}".format(err)
                sys.stdout.write(msg)
                sys.stdout.write("  database:   {}\n".format(database_from))
                sys.stdout.write("  retstart:   {}\n".format(start))
                # sys.stdout.write("  retmax:     {}\n".format(retmax))
                sys.stdout.write("  batch_size: {}\n".format(self.batch_size))
                sys.stdout.write("  linkname:   {}\n".format(linkname))
                sys.stdout.write("  webenv:     {}\n".format(webenv))
                sys.stdout.write("  querykey:   {}\n".format(querykey))

            if record is not None:
                try:
                    # Read the downloaded data from the socket handle
                    mtch = re.search(r'(ERROR.*\n)', record)
                    if mtch:
                        sys.stdout.write(mtch.group(1))
                    # ostrm.write(record)
                    # ostrm.flush()
                # pylint: disable=broad-except
                except Exception as err:
                    sys.stdout.write("*FATAL: BAD READ SOCKET HANDLE: {}\n".format(str(err)))
            else:
                sys.stdout.write("*FATAL: NO SOCKET HANDLE TO READ FROM\n")


# Copyright (C) 2016-present DV Klopfenstein. All rights reserved.



"""Download PubMed data to get PMID. Print NIH iCite summary using PMID.

     1) Download PubMed summary to get PMID, given DOI in bib entry
     2) Print PMID downloaded from PubMed and citekey from pubs
     3) Print NIH's iCite summary for each publication

"""

__copyright__ = "Copyright (C) 2019-present, DV Klopfenstein. All rights reserved."
__author__ = "DV Klopfenstein"

import os
import sys
from pmidcite.eutils.pubmed.rdwr import PubMedRdWr
from pmidcite.eutils.pubmed.record import PubMedRecord
from pmidcite.eutils.cmds.pubmed import PubMed


# pylint: disable=too-few-public-methods
class PubMedQuery:
    """Download PubMed text summary for PMID. Print NIH iCite summary using PMID"""

    patterns = {
        'doi' : '{doi}[Location ID] OR {doi}[Secondary Source ID] OR {doi}[Article Identifier]',
        'pmid' : '{pmid}[PMID]',
    }

    def __init__(self, email, apikey, tool, prt=sys.stdout):
        self.pubmed = PubMed(email, apikey, tool, prt)
        self.pubmedrw = PubMedRdWr()

    def get_pubmed_g_doi(self, doi, prt=None):
        """Give a DOI, return PubMed record's text and Record object"""
        query = self.patterns['doi'].format(doi=doi)
        pubmed_txt = self.dnld_text_g_query(query)
        if pubmed_txt is not None:
            pmid2dct = self.get_pmid2dct_g_txt(pubmed_txt, prt)
            pmid2rec = {pmid:PubMedRecord(dct) for pmid, dct in pmid2dct.items()}
            return {'pubmed_text': pubmed_txt, 'pmid2rec':pmid2rec}
        return {}

    def get_pubmed_g_pmid(self, pmid, prt=None):
        """Give a DOI, return PubMed record's text and Record object"""
        query = self.patterns['pmid'].format(pmid=pmid)
        pubmed_txt = self.dnld_text_g_query(query)
        if pubmed_txt is not None:
            pmid2dct = self.get_pmid2dct_g_txt(pubmed_txt, prt)
            pmid2rec = {pmid:PubMedRecord(dct) for pmid, dct in pmid2dct.items()}
            return {'pubmed_text': pubmed_txt, 'pmid2rec':pmid2rec}
        return {}

    def get_prt(self):
        """Return the print stored in the E-Utils base"""
        return self.pubmed.log

    def get_pubmed_dct(self, file_txt, prt=None, **kws):
        """Download or load one PubMed text summary as a dict for one publication"""
        if os.path.exists(file_txt):
            return self._get_pmiddct(file_txt)
        for key in set(self.patterns).intersection(kws):
            query = self.patterns[key].format(**{key:kws[key]})
            pubmed_txt = self.dnld_text_g_query(query)
            if pubmed_txt is not None:
                self.wr_text(file_txt, pubmed_txt)
                return self.get_pmid2dct_g_txt(pubmed_txt, prt)
            print('**WARNING: NO RESULTS FOUND FOR: {Q}'.format(Q=query))
            return {}
        raise RuntimeError('UNKNOWN PubMed DATA: {KWS}'.format(KWS=str(kws)))

    def dnld_text_g_query(self, query):
        """Get PubMed text, given a PubMed query"""
        ## print('QUERY: {Q}'.format(Q=query))
        pmids = self.pubmed.dnld_query_pmids(query)
        if not pmids:
            return None
        efetch_idxs, efetch_params = self.pubmed.epost_ids(
            pmids, 'pubmed', 10, 10000, rettype='medline', retmode='text')
        txts = self.pubmed.dnld_texts(efetch_idxs, efetch_params)
        return '\n'.join(txts)

    def get_pmid2dct_g_txt(self, pubmed_txt, prt=None):
        """Given text containing downloaded PubMed records"""
        return self.pubmedrw.get_pmid2info_g_textblock(pubmed_txt, prt=prt)

    @staticmethod
    def wr_text(fout_txt, pubmed_txt):
        """Write PubMed record text into a file"""
        with open(fout_txt, 'w') as prt:
            prt.write(pubmed_txt)
            print('  WROTE: {TXT}'.format(TXT=fout_txt))

    def _get_pmiddct(self, fin_txt):
        """Read PubMed text summary. Return PubMed dict, which contains PMID"""
        dct = self.pubmedrw.get_pmid2info_g_text(fin_txt)
        assert len(dct) == 1, dct
        return list(dct.values())[0]


# Copyright (C) 2019-present DV Klopfenstein. All rights reserved.



"""Access data in a PubMed record"""

__copyright__ = "Copyright (C) 2019-present, DV Klopfenstein. All rights reserved."
__author__ = "DV Klopfenstein"


class PubMedRecord:
    """Access data in a PubMed record"""

    def __init__(self, pubmed_dict):
        self.pmdct = pubmed_dict
        self.pmid = self.pmdct['PMID']
        self.doi = self._init_doi()

    def get_pmc(self):
        """Get the PubMed Central (PMC) ID, if it exists"""
        return self.pmdct.get('PMC')

    def get_bibtex(self, eprint=True):
        """Get bibtex fields and values for PMID and PMC"""
        # https://tex.stackexchange.com/questions/155532/biblatex-and-pubmed-pubmed-central-ids
        pmdct = self.pmdct
        pmid = pmdct['PMID']
        dct = {'pmid':pmid, 'doi':self.doi}
        if 'PMC' in pmdct:
            pmcid = pmdct['PMC']
            dct['pmcid'] = pmcid
            if eprint:
                dct['eprint'] = pmcid
                dct['eprinttype'] = 'pmcid'
            return dct
        if eprint:
            dct['eprint'] = pmid
            dct['eprinttype'] = 'pubmed'
        return dct

    def _init_doi(self):
        """Extract DOI from PubMed record"""
        dct = self.pmdct
        for aid in ['AID', 'LID']:
            if aid in dct:
                for typ, val in dct[aid].items():
                    if typ == 'doi':
                        return val
        return None


# Copyright (C) 2019-present DV Klopfenstein. All rights reserved.



"""Author object."""

__author__ = 'DV Klopfenstein'
__copyright__ = "Copyright (C) 2019-present DV Klopfenstein. All rights reserved."


class Author(object):
  """Hold list of Authors."""

  # pylint: disable=invalid-name
  def __init__(self, full_author_name):
    self.fau = full_author_name
    self.au = None
    self.ads = []

  def __str__(self):
    lst = ["\nFAU: {AU}".format(AU=self.fau), "AU:  {AU}".format(AU=self.au)]
    for affil in self.ads:
      lst.append("AD:  {AD}".format(AD=affil))
    return "\n".join(lst)


# Copyright (C) 2019-present DV Klopfenstein.  All rights reserved.



"""Given PMIDs, download PubMed records containing"""

__author__ = 'DV Klopfenstein'
__copyright__ = "Copyright (C) 2016-present DV Klopfenstein. All rights reserved."
__license__ = "GPL"

from os import system
from os import getcwd
from os.path import exists
import sys
from pmidcite.eutils.cmds.base import EntrezUtilities
from pmidcite.eutils.cmds.esearch import ESearch


class PubMed(EntrezUtilities):
    """Fetch and write text"""

    medline_text = {
        'rettype': 'medline',
        'retmode': 'text',
    }
    pat = 'PMIDs/epost={P} PMIDs/efetch={F} querykey({Q} of {Qmax}) start({S})'

    def __init__(self, email, apikey, tool, prt=sys.stdout):
        super(PubMed, self).__init__(email, apikey, tool, prt)
        self.esearch = ESearch(email, apikey, tool, prt)

    def dnld_query_pmids(self, query, num_ids_p_epost=10):
        """Searches an PubMed for a user query, writes resulting entries into one file."""
        return self.esearch.dnld_query_pmids(query, 'pubmed', num_ids_p_epost)

    def _dnld_wr_all(self, fout_pubmed, efetch_idxs, efetch_params):
        """Download and write all PMIDs PubMed text entries into one file"""
        if exists(fout_pubmed):
            system('rm {FILE}'.format(FILE=fout_pubmed))
        if not efetch_idxs:
            return
        for desc, start, pmids_exp, querykey in efetch_idxs:
            rsp_txt = self._run_efetch('pubmed', start, querykey, pmids_exp, desc, **efetch_params)
            if rsp_txt is not None:
                with open(fout_pubmed, 'a') as prt:
                    prt.write(rsp_txt)
                    prt.flush()
        print('  WROTE: {FILE}'.format(FILE=fout_pubmed))

    def _dnld_query(self, query):
        """Searches an NCBI database for a user search term, returns NCBI IDs."""
        dct = self.run_eutilscmd('esearch', db='pubmed', term=query, retmode='json')
        return dct

    def dnld_wr1_per_pmid(self, pmids, force_download, dir_pubmed_txt, pmid2name=None):
        """Download and write one PubMed text file entry per PMID"""
        if not exists(dir_pubmed_txt):
            dir_pubmed_txt = self._err_exists(dir_pubmed_txt)
        pmid_nt_list = self.get_pmid_nt_list(pmids, force_download, dir_pubmed_txt, pmid2name)
        efetch_idxs, efetch_params = self.epost_ids(pmids, 'pubmed', 10, 1, **self.medline_text)
        return self.esearch.dnld_wr1_per_id('pubmed', efetch_idxs, efetch_params, pmid_nt_list)

    @staticmethod
    def _err_exists(dir_pubmed_txt):
        """Print warning message if dir not exist, return current working directory"""
        cwd = getcwd()
        print('**WARNING: DIR({DIR_NAME}) NOT EXIST RETURNING CWD({CWD})'.format(
            DIR_NAME=dir_pubmed_txt,
            CWD=getcwd()))
        return cwd

    def dnld_texts(self, efetch_idxs, efetch_params):
        """Download and save one PMID PubMed entry into a text string"""
        txts = []
        for desc, start, pmids_exp, querykey in efetch_idxs:
            rsp_txt = self._run_efetch('pubmed', start, querykey, pmids_exp, desc, **efetch_params)
            txts.append(rsp_txt)
        return txts

    @staticmethod
    def _get_str_post_fetch(num_pmids_p_epost, num_pmids_p_efetch, querykey, start):
        """Get a string summarizing current EPost and EFetch"""
        return 'IDs/epost={P} IDs/efetch={F} querykey({Q}) start({S})'.format(
            P=num_pmids_p_epost, F=num_pmids_p_efetch, Q=querykey, S=start)

    def get_pmid_nt_list(self, pmids, force_download, dir_pubmed, pmid2name=None):
        """Get list of PubMed entries: Title, abstract, authors, journal, MeSH"""
        # [Nt(PMID=31614060, file_pubmed='./log/pubmed/pubmed_31614060.txt', file_exists=False)]
        return self.esearch.get_pmid_nt_list(pmids, 'pubmed', force_download, dir_pubmed, pmid2name)



# Copyright (C) 2016-present DV Klopfenstein. All rights reserved.



"""Manage pmidcite Configuration"""

__copyright__ = "Copyright (C) 2019-present, DV Klopfenstein. All rights reserved."
__author__ = "DV Klopfenstein"

from os import environ
from os import getcwd
from os.path import exists
from os.path import basename
from os.path import join
from os.path import abspath
from sys import stdout
import configparser

from pmidcite.icite.nih_grouper import NihGrouper


def get_cfgparser(prt=stdout):
    """Init cfg parser"""
    cfgparser = Cfg(check=False, prt=prt)
    cfgparser.rd_rc(prt=prt)
    return cfgparser

# pylint: disable=useless-object-inheritance
class Cfg(object):
    """Manage pmidcite Configuration"""

    envvar = 'PMIDCITECONF'
    dfltcfgfile = '.pmidciterc'

    dfltdct = {
        'pmidcite' : {
            # NIH Group divisions
            'group1_min': '2.1',
            'group2_min': '15.7',
            'group3_min': '83.9',
            'group4_min': '97.5',

            # Entrez utilities
            'email': 'name@university.edu',
            # https://ncbiinsights.ncbi.nlm.nih.gov/2017/11/02/new-api-keys-for-the-e-utilities/
            'apikey': 'LONG_HEX_NCBI_API_KEY',
            'tool': 'scripts',

            # Information downloaded from NIH iCite stored in a Python module
            'dir_icite_py': 'None',

            # Directory for abstracts downloaded from PubMed
            'dir_pubmed_txt': 'None',

            # Used by PubMedQueryToICite:
            # Directory for files containing PMIDs downloaded from PubMed
            'dir_pmids': 'None',
            # Run NIH's iCite on a set of PMIDs and store results in a file
            'dir_icite': 'None',
        },
    }

    def __init__(self, check=True, prt=stdout, prt_fullname=True):
        self.cfgfile = self._init_cfgfilename()
        self.cfgparser = self._get_dflt_cfgparser()
        if check:
            self._run_chk(prt, prt_fullname)

    def prt_cfgfile(self, prt=stdout):
        """Print information about the configuration file"""
        prt.write('  printenv {VAR} # value({VAL})\n'.format(
            VAR=self.envvar,
            VAL=environ[self.envvar]))
        prt.write('  CFG FILE: {CFG}\n'.format(
            CFG=abspath(self.cfgfile)))

    def get_email(self):
        """Get email"""
        return self.cfgparser['pmidcite']['email']

    def get_apikey(self):
        """Get API Key"""
        return self.cfgparser['pmidcite']['apikey']

    def get_tool(self):
        """Get tool name"""
        return self.cfgparser['pmidcite']['tool']

    def get_dir_pubmed_txt(self):
        """Get the name of the directory containg PubMed entry text files"""
        return self.cfgparser['pmidcite']['dir_pubmed_txt']

    def get_dir_icite_py(self):
        """Get the name of the directory containg PubMed entry text files"""
        return self.cfgparser['pmidcite']['dir_icite_py']

    def get_dir_icite(self):
        """Get the name of the directory containg PubMed entry text files"""
        return self.cfgparser['pmidcite']['dir_icite']

    def get_fullname_pmids(self, filename):
        """get the name of the directory containg pubmed entry text files"""
        return self._get_fullname(filename, 'dir_pmids')

    def get_fullname_icite(self, filename):
        """get the name of the directory containg pubmed entry text files"""
        return self._get_fullname(filename, 'dir_icite')

    def _get_fullname(self, filename, dir_key):
        """Get the filename, including optionally the directory and filename"""
        dirname = self.cfgparser['pmidcite'][dir_key]
        if dirname is None or dirname == 'None':
            return filename
        if dirname is not None and exists(dirname):
            return join(dirname, filename)
        cwd = getcwd()
        print('**WARNING: DIR NOT EXIST {DIR_TYPE}({DIR_NAME}) RETURNING CWD({CWD})'.format(
            DIR_TYPE=dir_key, DIR_NAME=dirname, CWD=cwd))
        return join(cwd, filename)

    def set_dir_pubmed_txt(self, dirname):
        """Set the name of the directory containg PubMed entry text files"""
        self.cfgparser['pmidcite']['dir_pubmed_txt'] = self._get_dirname_str(dirname)

    def set_dir_icite_py(self, dirname):
        """Set the name of the directory containg PubMed entry text files"""
        self.cfgparser['pmidcite']['dir_icite_py'] = self._get_dirname_str(dirname)

    def set_dir_icite(self, dirname):
        """Set the name of the directory containg PubMed entry text files"""
        self.cfgparser['pmidcite']['dir_icite'] = self._get_dirname_str(dirname)

    def set_dir_pmids(self, dirname):
        """Set the name of the directory containg PubMed entry text files"""
        self.cfgparser['pmidcite']['dir_pmids'] = self._get_dirname_str(dirname)

    @staticmethod
    def _get_dirname_str(dirname):
        """Convert None to the str, "None", as needed by configparser"""
        return 'None' if dirname is None or dirname == 'None' else dirname

    def get_nihgrouper(self):
        """Get an NIH Grouper with default values from the cfg file"""
        cfg = self.cfgparser['pmidcite']
        return NihGrouper(
            float(cfg['group1_min']),
            float(cfg['group2_min']),
            float(cfg['group3_min']),
            float(cfg['group4_min']))

    def _run_chk(self, prt, prt_fullname):
        if not self.rd_rc(prt, prt_fullname):
            self._err_notfound()
        dflt = self.cfgparser['pmidcite']
        self._chk_email(dflt)
        self._chk_apikey(dflt)

    def set_cfg(self, cfgfile=None):
        """Set config file and initialize ConfigParser()"""
        self.cfgfile = self.dfltcfgfile if cfgfile is None else cfgfile
        self.cfgparser = self._get_dflt_cfgparser()
        return self.cfgparser.read(self.cfgfile)

    def rd_rc(self, prt=stdout, prt_fullname=True):
        """Read a configuration file"""
        if self.cfgfile is not None and exists(self.cfgfile):
            if prt:
                cfgfile = self.cfgfile if prt_fullname else basename(self.cfgfile)
                prt.write('  READ: {CFG}\n'.format(CFG=cfgfile))
        # Returns a list containing configuration file names
        return self.cfgparser.read(self.cfgfile)

    def wr_rc(self, force=False):
        """Write a sample configuration with default values set"""
        if not exists(self.cfgfile) or force:
            with open(self.cfgfile, 'w') as prt:
                self.cfgparser.write(prt)
                print('  WROTE: {CFG}'.format(CFG=self.cfgfile))
                return True
        print('  EXISTS: {CFG} OVERWRITE WITH wr_rc(force=True)'.format(CFG=self.cfgfile))
        return False

    def _chk_email(self, loaded):
        """Check to see that user has added their email"""
        if loaded['email'] == self.dfltdct['pmidcite']['email']:
            raise RuntimeError('SET EMAIL IN {CFG}'.format(CFG=self.cfgfile))

    def _chk_apikey(self, loaded):
        """Check to see that user has added a NCBI API key"""
        try:
            int(loaded['apikey'], 16)
        except ValueError:
            msg = ('SET API KEY IN {CFG}\n'
                   'Get an NCBI API key to run the E-utilities:\n'
                   'https://ncbiinsights.ncbi.nlm.nih.gov/2017/11/02/'
                   'new-api-keys-for-the-e-utilities\n'
                   'To ensure your API key is not made public, add {CFG} to the .gitignore')
            raise RuntimeError(msg.format(CFG=self.cfgfile))

    def _err_notfound(self):
        """Report the config file was not found"""
        cfgfile = environ[self.envvar] if self.envvar in environ else self.cfgfile
        msg = ('E-Utils CONFIG FILE NOT FOUND: {CFG}\n'
               'Generate {CFG} with:\n    '
               "$ python3 -c 'from pmidcite.cfg import Cfg; Cfg(check=False).wr_rc(force=True)'\n"
               'To ensure your API key is not made public, add {CFG} to the .gitignore')
        raise RuntimeError(msg.format(CFG=cfgfile))

    def _get_dflt_cfgparser(self):
        """Create a ConfigParser() filled with the default key-value"""
        config = configparser.ConfigParser()
        for section, dfltdct_cur in self.dfltdct.items():
            ## print('KEY-VAL: {} {}'.format(section, dfltdct_cur))
            config[section] = dfltdct_cur
        return config

    def prt_rcfile_dflt(self, prt=stdout):
        """Print default rcfile"""
        cfgparser = self._get_dflt_cfgparser()
        cfgparser.write(prt)

    def _init_cfgfilename(self):
        """Get the configuration filename"""
        if self.envvar in environ:
            cfgfile = environ[self.envvar]
            if exists(cfgfile):
                return cfgfile
            print('**WARNING: NO pmidcite CONFIG FILE FOUND AT {ENVVAR}={F}'.format(
                F=cfgfile, ENVVAR=self.envvar))
        if not exists(self.dfltcfgfile):
            print('**WARNING: NO pmidcite CONFIG FILE FOUND: {F}'.format(
                F=self.dfltcfgfile))
        return self.dfltcfgfile



# Copyright (C) 2019-present DV Klopfenstein. All rights reserved.



"""ASCII scatter plot adapted from https://github.com/dzerbino/ascii_plots"""

__copyright__ = "Copyright (C) 2013-present, Authored by Daniel Zerbino. All rights reserved."
__copyright__ = "Copyright (C) 2020-present, Adapted by DV Klopfenstein. All rights reserved."

from itertools import chain
import sys
from collections import Counter
import locale
locale.setlocale(locale.LC_ALL, '')


class AsciiScatter:
    """ASCII scatter plot adapted from https://github.com/dzerbino/ascii_plots"""

    scale = r' 123456789abcdef!@#$%^&*(ABCDEFGHIJKLMNOPQRSTUVWXYZ-=_+{}[]|\:";<>,./?'

    def __init__(self, maxwidth=100):
        self.maxwidth = maxwidth

    def run(self, xydata):
        """Create an ASCII plot, given XY data"""
        minmax = self._get_minmax(xydata)
        cnts_2d = self._get_xy_scaled(xydata, **minmax)
        letter2cnt = self._asciifyarray(cnts_2d, **minmax)
        self._prt_key(letter2cnt)

    @staticmethod
    def read_stdin_ints():
        """Get (x, y) points from stdin"""
        xydata = []
        for line in sys.stdin:
            lst = line.split()
            if len(lst) == 2 and lst[0].isdigit() and lst[1].isdigit():
                xydata.append((int(lst[0]), int(lst[1])))
        return xydata

    def _asciify_2d(self, array_2d, max_val):
        chrs_2d = []
        letter2cnt = {}
        scale = self.scale
        scale_range = len(self.scale) - 1
        for array_1d in array_2d:
            chrs_1d = []
            for cnt in array_1d:
                scale_idx = cnt%scale_range
                letter = scale[scale_idx]
                letter2cnt[letter] = cnt
                chrs_1d.append(letter)
            chrs_2d.append(chrs_1d)
        return {'chrs_2d':chrs_2d, 'letter2cnt':letter2cnt}

    def _get_width(self, span_x):
        """Given the time span for citations, return width of ASCII graph"""
        multiplier = self.maxwidth//span_x
        return span_x*multiplier

    def _asciifyarray(self, array_2d, width, **kws):
        min_x, max_x, min_y, max_y, span_x = [kws[k] for k in ['min_x', 'max_x', 'min_y', 'max_y', 'span_x']]
        max_val = max(chain(*array_2d))
        print("".join("-" for i in range(width + 2)) + " " + str(max_y))
        dct = self._asciify_2d(array_2d, max_val)
        for chrs_1d in dct['chrs_2d']:
            print('|' + ''.join(chrs_1d) + "|")
        print("".join("-" for i in range(width + 2)) + " " + str(min_y))
        print(str(min_x) + " "*width + str(max_x))
        return dct['letter2cnt']

    def _prt_key(self, letter2cnt, prt=sys.stdout):
        """Print the counts and the letter that represents it"""
        for letter, cnt in sorted(letter2cnt.items(), key=lambda t: t[1]):
            if cnt != 0 and letter != str(cnt):
                prt.write('{A} {N}'.format(A=letter, N=cnt))

    def _get_xy_scaled(self, xydata, width, **kws):
        """Scale XY values to fit in ASCII width"""
        min_x, max_y, span_x, span_x, span_y = [kws[k] for k in ['min_x', 'max_y', 'span_x', 'span_x', 'span_y']]
        array = [[0]*width for i in range(width//3)]
        for xval, yval in sorted(xydata, key=lambda t: t[0]):
            pos_x = int(round((width - 1) * (xval - min_x) / span_x))
            pos_y = int(round((width//3 - 1) * (max_y - yval) / span_y))
            array[pos_y][pos_x] += 1
            ## print('X({}) Y({:3}) -> {} {:2} VALUE {}'.format(xval, yval, pos_x, pos_y, array[pos_y][pos_x]))
        return array

    def _get_minmax(self, xydata):
        """Get min and max data from XY xydata"""
        if not xydata:
            return {}
        xvals, yvals = zip(*xydata)
        min_x = min(xvals)
        max_x = max(xvals)
        min_y = min(yvals)
        max_y = max(yvals)
        span_x = max_x - min_x
        return {
            'min_x': min_x, 'max_x': max_x, 'span_x': span_x,
            'min_y': min_y, 'max_y': max_y, 'span_y': max_y - min_y,
            'width': self._get_width(span_x),
        }


# Copyright (C) 2013-present, Authored by Daniel Zerbino. All rights reserved.
# Copyright (C) 2020-present, Adapted for pmidcite by DV Klopfenstein. All rights reserved.



"""Fetch items and write"""
# https://www.ncbi.nlm.nih.gov/books/NBK25499/#chapter4.EFetch

__author__ = 'DV Klopfenstein'
__copyright__ = "Copyright (C) 2016-present DV Klopfenstein. All rights reserved."
__license__ = "GPL"

import sys
import re
from pmidcite.eutils.cmds.cmdbase import CommandBase


class EFetch(CommandBase):
    """Fetch and write text"""

    # pylint: disable=too-many-arguments
    #### def __init__(self, retmax=10000, rettype='medline', retmode='text', batch_size=100, **kws):
    def __init__(self, rettype='medline', retmode='text', batch_size=100, **kws):
        kws_base = {k:v for k, v in kws.items() if k in CommandBase.exp_kws}
        print('FFFFFFFFFFFFFFFFFFFF', kws_base)
        super(EFetch, self).__init__(**kws_base)

    def efetch_and_write(self, ostrm, database, webenv, querykey, num_fetches):
        """EFetch records found for PMIIDs, page by page"""
        ## QueryKey(     1) EFetching(database=pubmed) up to    10 records, starting at 0; ABSTRACT
        ## QueryKey(     1) EFetching(database=pubmed) up to    10 records, starting at 10; ABSTRACT
        ## msg_fmt = ('  QueryKey({:>6}) EFetching(database={}) up to {:5} records, '
        ##            'starting at {}; {}\n')
        for start in range(0, num_fetches, self.batch_size):
            ## msg = msg_fmt.format(querykey, database, self.batch_size, start, self.desc)
            ## sys.stdout.write(msg)
            print('SSSSSSSSSSSSSSSSSSSSSSSTART:', start)
            txt = self.efetch_txt(start, self.batch_size, database, webenv, querykey)

            if txt is not None:
                try:
                    # Read the downloaded data from the socket handle
                    mtch = re.search(r'(ERROR.*\n)', txt)
                    if mtch:
                        sys.stdout.write(mtch.group(1))
                    ostrm.write(txt)
                    ostrm.flush()
                # pylint: disable=broad-except
                except Exception as err:
                    sys.stdout.write("*FATAL: BAD READ SOCKET HANDLE: {}\n".format(str(err)))
            else:
                sys.stdout.write("*FATAL: NO SOCKET HANDLE TO READ FROM\n")

    def efetch_txt(self, start, retmax, database, webenv, querykey):
        """Fetch database text"""
        try:
            # pylint: disable=bad-whitespace
            txt = self.run_eutilscmd(
                'efetch',
                db        = database,
                retstart  = start,         # dflt: 1
                retmax    = retmax,        # max: 10,000
                rettype   = self.rettype,  # Ex: medline
                retmode   = self.retmode,  # Ex: text
                webenv    = webenv,
                query_key = querykey)
            #print('FETCH:', dct)
            return txt
        except IOError as err:
            msg = "\n*FATAL: EFetching FAILED: {}".format(err)
            sys.stdout.write(msg)
            sys.stdout.write("  database:   {}\n".format(database))
            sys.stdout.write("  retstart:   {}\n".format(start))
            sys.stdout.write("  batch_size: {}\n".format(self.batch_size))
            sys.stdout.write("  rettype:    {}\n".format(self.rettype))
            sys.stdout.write("  retmode:    {}\n".format(self.retmode))
            sys.stdout.write("  webenv:     {}\n".format(webenv))
            sys.stdout.write("  querykey:   {}\n".format(querykey))


# Copyright (C) 2016-present DV Klopfenstein. All rights reserved.



#!/usr/bin/env python3
"""Read a file created by pmidcite and write simple text file of PMIDs"""

__copyright__ = "Copyright (C) 2019-present, DV Klopfenstein. All rights reserved."
__author__ = "DV Klopfenstein"


from pmidcite.cli.readpmids import ReadPmids  # get_argparser


def main():
    """Print lists of pubs in formation"""
    ReadPmids().cli()


if __name__ == '__main__':
    main()

# Copyright (C) 2019-present, DV Klopfenstein. All rights reserved.



"""Given a PubMed ID (PMID), return a list of publications which cite it"""

__copyright__ = "Copyright (C) 2019-present, DV Klopfenstein. All rights reserved."
__author__ = "DV Klopfenstein"

from sys import stdout
from os.path import join
from os.path import exists
from importlib.util import spec_from_file_location
from importlib.util import module_from_spec

from pmidcite.icite.entry import NIHiCiteEntry


class NIHiCiteLoader:
    """Load iCite citations that are stored as a dict in a Python module"""

    def __init__(self, nih_grouper, dir_icitepy, assc_pmid_keysset, icitepypat='p{PMID}.py'):
        self.nih_grouper = nih_grouper
        self.dir_dnld = dir_icitepy  # e.g., ./icite
        self.associated_pmid_keysset = assc_pmid_keysset
        self.icitepypat = icitepypat

    def load_icites(self, pmids, prt=stdout):
        """Load multiple NIH iCite data from Python modules"""
        if not pmids:
            return []
        icites = []
        s_icitepat = self.icitepypat
        s_load_icite = self.load_icite
        for pmid in pmids:
            file_pmid = join(self.dir_dnld, s_icitepat.format(PMID=pmid))
            iciteobj = s_load_icite(file_pmid)
            if iciteobj is not None:
                icites.append(iciteobj)
        if prt:
            num_icites = len(icites)
            num_pmids = len(pmids)
            if num_icites != num_pmids:
                prt.write('{N:5,} of {P:5,} PMIDs have iCite entries\n'.format(
                    N=num_icites,
                    P=num_pmids))
        return icites

    def load_icite_mods_all(self, pmids_top):
        """Load NIHiCiteEntry for the citations and references of PMIDs in pmids_top"""
        icites_top = self.load_icites(pmids_top)                             # [NIHiCiteEntry]
        pmids_top = set(o.dct['pmid'] for o in icites_top)
        pmids_linked = self._get_pmids_linked(icites_top)                    # [NIHiCiteEntry]
        icites_linked = self.load_icites(pmids_linked.difference(pmids_top)) # [NIHiCiteEntry]
        ## print('LLLLLLLLLL NIHiCiteLoader load_icite_mods_all icites_top', icites_top)
        ## print('LLLLLLLLLL NIHiCiteLoader load_icite_mods_all pmids_linked', len(pmids_linked))
        ## print('LLLLLLLLLL NIHiCiteLoader load_icite_mods_all icites_linked', len(icites_linked))
        return icites_top + icites_linked

    def get_file_pmid(self, pmid):
        """Get the name of the icite file for one PMID"""
        return join(self.dir_dnld, self.icitepypat.format(PMID=pmid))

    def load_icite(self, file_pmid):
        """Load NIH iCite information from Python modules"""
        if exists(file_pmid):
            spec = spec_from_file_location("module.name", file_pmid)
            mod = module_from_spec(spec)
            spec.loader.exec_module(mod)
            ## print('LLLLLLLLLLLLL load_icite', file_pmid)
            return NIHiCiteEntry(mod.ICITE, self.nih_grouper.get_group(mod.ICITE['nih_percentile']))
        return None

    def load_pmid(self, pmid):
        """Get NIHiCiteEntry for a PMID"""
        fin_py = self.get_file_pmid(pmid)
        return self.load_icite(fin_py)

    def _get_pmids_linked(self, icites_top):
        """Get the PMIDs for the citations and references of top NIHiCiteEntry"""
        s_asscpmid_keys = self.associated_pmid_keysset
        return set(pmid for o in icites_top for f in s_asscpmid_keys for pmid in o.dct[f])
        ## pmids_linked = set()
        ## for obj in icites_top:
        ##     for fld in self.associated_pmid_keysset:
        ##         for pmid in obj.dct[fld]:
        ##             pmids_linked.add(pmid)
        ## return pmids_linked


# Copyright (C) 2019-present DV Klopfenstein. All rights reserved.



"""Manage counts for various types of PubMed data"""

__copyright__ = "Copyright (C) 2019-present, DV Klopfenstein. All rights reserved."
__author__ = "DV Klopfenstein"

import collections as cx

# Notes from on non-MEDLINE citations accessible through PubMed:
#     McKeever, Liam et al.
#    "Demystifying the Search Button"
#     Journal of Parenteral and Enteral Nutrition (2015)
#
# Of the remaining 11.1%:
#     * 23.3% are in the process of becoming MEDLINE indexed,
#     * 15.3% have been submitted by the publisher and may become MEDLINE indexed in the future,
#     * 51.7% are available on PubMed but will not be indexed by MEDLINE, and
#     * 9.7% belong to an older version of MEDLINE

# pylint: disable=line-too-long
class DataMgr:
    """Manage counts for various types of PubMed data"""

    ntobj = cx.namedtuple('Nt', 'count perc')

    def __init__(self, abc2count):
        self.a2n = abc2count
        # All of PubMed
        self.tot_pubmed = abc2count['all']
        # MEDLINE, PMC, and inprocess
        self.tot_pmc = abc2count['ml1_pmc1']
        self.pltdata_pubmed = self._init_plot_data_pubmed()

    def _init_plot_data_pubmed(self):
        """Get data for the PubMed contents plot"""
        # pylint: disable=bad-whitespace
        name_data = []
        a2n = self.a2n
        #
        name_data.append(('PubMed',              self._get_nt_pubmed(a2n['all'])))
        name_data.append(('MEDLINE_and_PMC',     self._get_nt_pubmed(a2n['ml1_pmc1'])))
        name_data.append(('MEDLINE_n_inprocess', self._get_nt_pubmed(a2n['medline_n_inprocess'])))
        name_data.append(('PMC',                 self._get_nt_pubmed(a2n['pmc_all'])))
        name_data.append(('PMC_ml',              self._get_nt_pubmed(a2n['pmc_m1'])))
        name_data.append(('PMC_only',            self._get_nt_pubmed(a2n['pmc_m0'])))
        name_data.append(('other',               self._get_nt_pubmed(a2n['all_ml0_pmc0'])))
        #
        name_data.append(('ip',         self._get_nt_pubmed(a2n['inprocess_A_all'])))
        name_data.append(('ip_pmc0',    self._get_nt_pubmed(a2n['inprocess_A_pmc0'])))
        name_data.append(('ip_pmc1',    self._get_nt_pubmed(a2n['inprocess_A_pmc1'])))
        #
        name_data.append(('pmc_ip1',    self._get_nt_pubmed(a2n['inprocess_A_pmc1'])))
        name_data.append(('pmc_ml1',    self._get_nt_pubmed(a2n['medline_pmc1'])))
        name_data.append(('pmc_ml0',    self._get_nt_pubmed(a2n['pmc_m0'])))
        #
        name_data.append(('ml_pmc0',     self._get_nt_pubmed(a2n['medline_pmc0'])))
        name_data.append(('ml_pmc0_ip1', self._get_nt_pubmed(a2n['inprocess_A_pmc0'])))
        name_data.append(('ml_pmc1_ip1', self._get_nt_pubmed(a2n['inprocess_A_pmc1'])))
        name_data.append(('ml_pmc1',     self._get_nt_pubmed(a2n['medline_pmc1'])))
        dct = cx.OrderedDict(name_data)
        # Check that counts for the PubMed content figure add up
        #                             BLUE|CYAN|BROWN              + ORANGE
        assert dct['PubMed'].count == dct['MEDLINE_and_PMC'].count + dct['other'].count
        #                             BLUE|CYAN                        + BROWN                 + ORANGE
        assert dct['PubMed'].count == dct['MEDLINE_n_inprocess'].count + dct['PMC_only'].count + dct['other'].count
        #                             CYAN
        assert dct['ip'].count == dct['ip_pmc0'].count + dct['ip_pmc1'].count
        #                             CYAN                BLUE                   BROWN
        assert dct['PMC'].count == dct['pmc_ip1'].count + dct['pmc_ml1'].count + dct['pmc_ml0'].count
        #                                          BLUE                   CYAN                       CYAN                   BROWN
        assert dct['MEDLINE_n_inprocess'].count == dct['ml_pmc0'].count + dct['ml_pmc0_ip1'].count + dct['ml_pmc1_ip1'].count + dct['ml_pmc1'].count
        #                          CYAN|BLUE             BROWN
        assert dct['PMC'].count == dct['PMC_ml'].count + dct['PMC_only'].count
        return dct

    def _get_nt_pubmed(self, cnt):
        """Return namedtuple with count and percentage data"""
        return self.ntobj(count=cnt, perc=100.0*(cnt/self.tot_pubmed))

    def get_pubmed_xvals(self):
        """Get the colorbar data for the PubMed plot"""
        a2n = self.pltdata_pubmed
        xvals = [a2n['ml_pmc0'].count]                      # 0
        xvals.append(xvals[-1] + a2n['ml_pmc0_ip1'].count)  # 1
        xvals.append(xvals[-1] + a2n['ml_pmc1_ip1'].count)  # 2
        xvals.append(xvals[-1] + a2n['ml_pmc1'].count)      # 3
        xvals.append(xvals[-1] + a2n['PMC_only'].count)     # 4
        # print(a2n['PubMed'], xvals[-1] + a2n['other'].count)
        assert a2n['PubMed'].count == xvals[-1] + a2n['other'].count
        return xvals

    def get_pubmed_colorbars(self, yval):
        """Get the colorbar data for the PubMed plot"""
        par = {'edgecolor': 'black', 'linewidth':0.0, 'alpha':1.0}
        xvals = self.get_pubmed_xvals()
        a2n = self.pltdata_pubmed
        # pylint: disable=bad-whitespace
        # pylint: disable=line-too-long
        return [
            # All PubMed
            ## ([(0, a2n['all'])],                ( 0, 2), {'facecolors':'k', **par}),
            # MEDLINE
            ([(0,        a2n['ml_pmc0'].count)], (yval-1, 1.8), {'label':'MEDLINE', 'facecolors':'tab:blue', **par}),
            ([(xvals[0], a2n['ip'].count)],      (yval-1, 1.8), {'label':'MEDLINE in process', 'facecolors':'lawngreen', **par}),
            ([(xvals[2], a2n['ml_pmc1'].count)], (yval-1, 1.8), {'facecolors':'tab:blue', **par}),
            # PMC
            ([(xvals[1], a2n['pmc_ip1'].count)], (yval-3, 1.8), {'facecolors':'lawngreen', **par}),
            ([(xvals[2], a2n['pmc_ml1'].count)], (yval-3, 1.8), {'facecolors':'tab:blue', **par}),
            ([(xvals[3], a2n['pmc_ml0'].count)], (yval-3, 1.8), {'label':'PMC Only', 'facecolors':'brown', **par}),
            # Other
            ([(xvals[4], a2n['other'].count)],   (yval-7, 1.8), {'label':'Other', 'facecolors':'tab:orange', **par}),
        ]

# Copyright (C) 2019-present, DV Klopfenstein. All rights reserved.



"""Text query finds database UIDs for later use in ESummary, EFetch or ELink"""
# https://www.ncbi.nlm.nih.gov/books/NBK25499/#chapter4.ESearch

__copyright__ = "Copyright (C) 2020-present DV Klopfenstein. All rights reserved."
__author__ = 'DV Klopfenstein'

import os
import sys
import collections as cx
from pmidcite.eutils.cmds.base import EntrezUtilities
from pmidcite.eutils.cmds.query_ids import QueryIDs


class ESearch(EntrezUtilities):
    """Text query finds database UIDs for later use in ESummary, EFetch or ELink"""

    #### # https://www.ncbi.nlm.nih.gov/books/NBK25499/#chapter4.ESearch
    #### exp_params = {
    ####     # Required Parameters:
    ####     'db',
    ####     'term',
    ####     # Optional Parameters - History Server
    ####     'usehistory',
    ####     'webenv',
    ####     'querykey',
    ####     # Optional Parameters – Retrieval
    ####     'retstart',
    ####     'retmax',
    ####     'rettype',
    ####     'retmode',
    ####     'sort',
    ####     'field',
    ####     'idtype',
    ####     # Optional Parameters – Dates
    ####     'datetype',
    ####     'reldate',
    ####     'mindate',
    ####     'maxdate',
    #### }
    #### pat = 'IDs/epost={P} IDs/efetch={F} querykey({Q} of {Qmax}) start({S})'


    def __init__(self, email, apikey, tool, prt=sys.stdout):
        super(ESearch, self).__init__(email, apikey, tool, prt)
        self.queryids = QueryIDs(email, apikey, tool, prt)

    def dnld_query_pmids(self, query, database, num_ids_p_epost=10):
        """Searches a NCBI database for a user query, writes resulting entries into one file."""
        return self.queryids.dnld_query_ids(query, database, num_ids_p_epost)
        #### # 1) Query PubMed/Protein/etc. Get first N (num_ids_p_epost) of the total PMIDs
        #### rsp_dct = self.query(database, query, retmax=num_ids_p_epost)
        #### if rsp_dct is None:
        ####     if self.log:
        ####         self.log.write('No {DB} entries found: {Q}\n'.format(DB=database, Q=query))
        ####         self.log.flush()
        ####     return []
        #### tot_ids = rsp_dct['count']
        #### ids = list(rsp_dct['idlist'])
        #### if rsp_dct and self.log:
        ####     self.log.write('{N:6,} IDs FOR {DB} QUERY({Q})\n'.format(
        ####         DB=database, N=tot_ids, Q=query))
        ####     self.log.flush()
        #### # 2) Continue to download PMIDs, N (num_ids_p_epost) at a time
        #### kws_p = {
        ####     'webenv': rsp_dct['webenv'],
        ####     'querykey': rsp_dct['querykey'],
        ####     'retmax': num_ids_p_epost,
        #### }
        #### for retnum in range(1, self._get_num_querykeys(num_ids_p_epost, tot_ids)):
        ####     rsp_dct = self.query(database, query, retstart=num_ids_p_epost*retnum, **kws_p)
        ####     if rsp_dct:
        ####         ids.extend(rsp_dct['idlist'])
        #### assert tot_ids == len(set(ids)), 'PMIDS EXP({E}) ACT({A})'.format(
        ####     E=tot_ids, A=len(set(ids)))
        #### return ids

    def dnld_wr1_per_id(self, database, efetch_idxs, efetch_params, pmid_nt_list):
        """Download and write one PMID PubMed entry into one text file"""
        if not pmid_nt_list:
            return {}
        pmid2nt = {nt.PMID:nt for nt in pmid_nt_list}
        for desc, start, pmids_exp, querykey_cur in efetch_idxs:
            rsp_txt = self._run_efetch(
                database, start, querykey_cur, pmids_exp, desc, **efetch_params)
            if rsp_txt is not None:
                assert len(pmids_exp) == 1
                pmid = pmids_exp[0]
                if pmid in pmid2nt:
                    ntd = pmid2nt[pmid]
                    if ntd.download:
                        ## print('NNNNNNNNNNNNNNN', ntd)
                        with open(ntd.file_pubmed, 'w') as prt:
                            prt.write(rsp_txt)
                            print('  {WROTE}: {TXT}'.format(
                                WROTE='WROTE' if not ntd.file_exists else 'UPDATED',
                                TXT=ntd.file_pubmed))
                    else:
                        print('**NOTE: NOT DOWNLOADING: {PMID}'.format(PMID=pmid))
        return pmid2nt

    #### @staticmethod
    #### def _get_num_querykeys(num_ids_p_epost, num_pmids):
    ####     """Get the number of querykeys necessary to process all PMIDs"""
    ####     num_querykeys = num_pmids//num_ids_p_epost
    ####     if num_pmids%num_ids_p_epost != 0:
    ####         num_querykeys += 1
    ####     return num_querykeys

    def esearch_ids(self, database, query, **return_params):
        """Get IDs using ESearch"""
        # Run query to find matching IDs. Get webenv for ID list
        rsp_0 = self.queryids.query(database, query, **return_params)
        idlist = list(rsp_0['idlist'])
        webenv = rsp_0['webenv']
        retmax = rsp_0['retmax']
        #self._prt_rsp(rsp_0, 0)
        num_iter = self._get_num_iterations(rsp_0['count'], rsp_0['retmax'])
        for querykey_cur in range(1, num_iter):
            rsp_i = self.queryids.query(
                database, query, WebEnv=webenv, retstart=querykey_cur*retmax, retmax=retmax)
            idlist.extend(rsp_i['idlist'])
            # self._prt_rsp(rsp_i, querykey_cur)
        print('{N} of {C} {DB} IDs FOR QUERY({Q})'.format(
            N=len(idlist), DB=database, C=rsp_0['count'], Q=query))
        ## print(len(set(idlist)))
        return idlist

    #### def query(self, database, query, **esearch):
    ####     """Text query finds database UIDs for later use in ESummary, EFetch or ELink"""
    ####     kws_exp = self.exp_params.difference(
    ###         {'db', 'term', 'rettype', 'usehistory', 'retmode'})
    ####     kws_act = {k:v for k, v in esearch.items() if k in kws_exp}
    ####     # Returns:
    ####     #    count
    ####     #    retmax
    ####     #    retstart
    ####     #    querykey
    ####     #    webenv
    ####     #    idlist
    ####     #    translationset
    ####     #    translationstack
    ####     #    querytranslation
    ####     dct = self.run_eutilscmd(
    ####         'esearch',
    ####         db=database,
    ####         term=query,
    ####         rettype='uilist',
    ####         # retmax=esearch.get('retmax', 10),
    ####         usehistory="y", # NCBI prefers we use history(QueryKey, WebEnv) for next acess
    ####         retmode='json',
    ####         **kws_act)
    ####     if dct is not None and 'idlist' in dct and dct['idlist']:
    ####         if database in {'pubmed',}:
    ####             dct['idlist'] = [int(n) for n in dct['idlist']]
    ####         for fldname in {'count', 'retmax'}:
    ####             dct[fldname] = int(dct[fldname])
    ####         return dct
    ####     return None

    @staticmethod
    def _prt_rsp(rsp_dct, idx):
        """Print key-value pairs in a response dict"""
        for key, val in rsp_dct.items():
            if key == 'idlist':
                print('{I} FFFFFFFFFFFF {K:20} {V}'.format(I=idx, K=key, V=len(val)))
            elif key not in 'translationstack':
                print('{I} FFFFFFFFFFFF {K:20} {V}'.format(I=idx, K=key, V=val))

    @staticmethod
    def get_pmid_nt_list(ids, database, force_download, dir_pubmed, pmid2name=None):
        """Get list of database entries. PubMed ex: Title, abstract, authors, journal, MeSH"""
        nts = []
        ntobj = cx.namedtuple('Nt', 'PMID file_pubmed file_exists download')
        for id_val in ids:
            name = pmid2name[id_val] if pmid2name and id_val in pmid2name else id_val
            # Get filename, pubmed_PMID.txt
            file_db = os.path.join(dir_pubmed, '{DB}_{ID}.txt'.format(DB=database, ID=name))
            file_exists = os.path.exists(file_db)
            download = not file_exists or force_download
            ntd = ntobj(
                PMID=id_val,
                file_pubmed=file_db,
                file_exists=file_exists,
                download=download)
            nts.append(ntd)
            if not download:
                print('**NOTE: EXISTS: {TXT}'.format(TXT=file_db))
        return nts


# Copyright (C) 2020-present, DV Klopfenstein. All rights reserved.



"""Manage args for NIH iCite run for one PubMed ID (PMID)"""

__copyright__ = "Copyright (C) 2019-present, DV Klopfenstein. All rights reserved."
__author__ = "DV Klopfenstein"

from sys import stdout
import argparse

from pmidcite.eutils.cmds.pubmed import PubMed
from pmidcite.cfgini import prt_rcfile
from pmidcite.cli.utils import get_outfile
from pmidcite.cli.utils import get_pmids
from pmidcite.cli.entry_keyset import get_details_cites_refs
from pmidcite.icite.nih_grouper import NihGrouper
from pmidcite.icite.downloader import get_downloader
from pmidcite.icite.downloader import prt_hdr
from pmidcite.icite.downloader import prt_keys


class NIHiCiteCli:
    """Manage args for NIH iCite run for one PubMed ID (PMID)"""

    def __init__(self, cfg):
        self.cfg = cfg
        self.pubmed = PubMed(
            email=cfg.get_email(),
            apikey=cfg.get_apikey(),
            tool=cfg.get_tool())

    def get_argparser(self):
        """Argument parser for Python wrapper of NIH's iCite given PubMed IDs"""
        parser = argparse.ArgumentParser(
            description="Run NIH's iCite given PubMed IDs",
            add_help=False)
        cfg = self.cfg
        dflt_dir_icite_py = cfg.get_dir_icite_py()
        dflt_dir_icite = cfg.get_dir_icite()
        dflt_dir_pubmed_txt = cfg.get_dir_pubmed_txt()
        # https://docs.python.org/3/library/argparse.html
        # https://docs.python.org/3/library/argparse.html#action
        # - PMIDs ----------------------------------------------------------------------------
        parser.add_argument(
            '-h', '--help', action='store_true',
            help='print this help message and exit (also --help)')
        parser.add_argument(
            'pmids', metavar='PMID', type=int, nargs='*',
            help='PubMed IDs (PMIDs)')
        parser.add_argument(
            '-i', '--infile', nargs='*',
            help='Read PMIDs from a file containing one PMID per line.')
        # - help -----------------------------------------------------------------------------
        parser.add_argument(
            '-H', '--print_header', action='store_true',
            help='Print column headings on one line.')
        parser.add_argument(
            '-k', '--print_keys', action='store_true',
            help='Print the keys describing the abbreviations.')
        # - verbosity ------------------------------------------------------------------------
        parser.add_argument(
            '-v', '--verbose', action='store_true', default=False,
            help='Load and print a descriptive list of citations and references for each paper.')
        parser.add_argument(
            '-c', '--load_citations', action='store_true', default=False,
            help='Load and print a descriptive list of citations for each paper.')
        parser.add_argument(
            '-r', '--load_references', action='store_true', default=False,
            help='Load and print a descriptive list of references for each paper.')
        # pylint: disable=line-too-long
        parser.add_argument(
            '-R', '--no_references', action='store_true',
            help='(DEPRECATED) Do not load or print a descriptive list of references. DEPRECATED -- Use instead: -c -r')
        # - output ---------------------------------------------------------------------------
        parser.add_argument(
            '-a', '--append_outfile',
            help='Append current citation report to an ASCII text file. Create if needed.')
        parser.add_argument(
            '-o', '--outfile',
            help='Write current citation report to an ASCII text file.')
        parser.add_argument(
            '-O', action='store_true',
            help="Write each PMIDs' iCite report with citations/references to <dir_icite>/PMID.txt")
        parser.add_argument(
            '-f', '--force_write', action='store_true',
            help='if an existing outfile file exists, overwrite it.')
        parser.add_argument(
            '-D', '--force_download', action='store_true',
            help='Download PMID iCite information to a Python file, over-writing if necessary.')
        # - abstracts -------------------------------------------------------------------------
        parser.add_argument(
            '-p', '--pubmed', action='store_true',
            help='Download PubMed entry containing title, abstract, authors, journal, MeSH, etc.')
        self.cfg.get_nihgrouper().add_arguments(parser)
        # - directories ----------------------------------------------------------------------
        # pylint: disable=line-too-long
        parser.add_argument(
            '--dir_icite_py', default=dflt_dir_icite_py,
            help='Write PMID iCite information into directory which contains temporary working files (default={D})'.format(
                D=dflt_dir_icite_py))
        parser.add_argument(
            '--dir_icite', default=dflt_dir_icite,
            help='Write PMID icite reports into directory (default={D})'.format(D=dflt_dir_icite))
        parser.add_argument(
            '--dir_pubmed_txt', default=dflt_dir_pubmed_txt,
            help='Write PubMed entry into directory (default={D})'.format(D=dflt_dir_pubmed_txt))
        # ------------------------------------------------------------------------------------
        parser.add_argument(
            '--md', action='store_true',
            help='Print using markdown table format.')
        # - initialization -------------------------------------------------------------------
        parser.add_argument(
            '--generate-rcfile', action='store_true',
            help='Generate a sample configuration file according to the '
                 'current configuration.')
        parser.add_argument(
            '--print-rcfile', action='store_true',
            help='Print the location of the pmidcite configuration file')
        return parser

    def cli(self):
        """Run iCite/PubMed using command-line interface"""
        argparser = self.get_argparser()
        args = self._get_args(argparser)
        ## print('ICITE ARGS ../pmidcite/src/pmidcite/cli/icite.py', args)
        self._run(args, argparser)

    def _run(self, args, argparser):
        """Run iCite/PubMed using command-line interface"""
        if args.help:
            argparser.print_help()
            print('\nHelp message printed because: -h or --help == True')
            exit()
        self.prt_info(args)
        # Get a list of researcher-specified PMIDs
        pmids = get_pmids(args.pmids, args.infile)
        if pmids:
            if len(pmids) > 10:
                print('PROCESSING {N:,} PMIDs'.format(N=len(pmids)))
            dnldr = self._get_downloader(args)
            pmid2icitepaper = dnldr.get_pmid2paper(pmids, None)
            ## print('XXXXXXXXXXXXXXXXXXXXXXXXXXXX pmid2icitepaper', pmid2icitepaper)
            self.run_icite(pmid2icitepaper, dnldr, args, argparser)
            if args.pubmed:
                self.pubmed.dnld_wr1_per_pmid(pmids, args.force_download, args.dir_pubmed_txt)
        # pylint: disable=line-too-long
        elif not (args.generate_rcfile or args.print_keys or args.print_header or args.print_rcfile):
            argparser.print_help(stdout)

    def prt_info(self, args):
        """Print helpful information"""
        # If writing to a file, don't write to stdout
        if args.O or args.append_outfile or args.outfile:
            return
        # Print rcfile initialization file
        if args.print_rcfile:
            self.cfg.prt_cfgfile()
        if args.generate_rcfile:
            prt_rcfile(stdout)
        self._prt_keys_n_hdr(stdout, args.print_keys, args.print_header)

    @staticmethod
    def _prt_keys_n_hdr(prt, print_keys, print_header):
        # Print the keys and/or the header
        if print_keys:
            prt_keys(prt)
        if print_header:
            prt_hdr(prt)

    @staticmethod
    def _get_downloader(args):
        """Get the downloader"""
        details_cites_refs = get_details_cites_refs(
            args.verbose,
            args.load_citations,
            args.load_references,
            args.no_references)
        groupobj = NihGrouper(args.min1, args.min2, args.min3, args.min4)
        return get_downloader(
            details_cites_refs,
            groupobj,
            args.dir_icite_py,
            args.force_download)

    def _get_args(self, argparser):
        """Get args"""
        args = argparser.parse_args()
        self.cfg.set_dir_icite_py(args.dir_icite_py)
        self.cfg.set_dir_icite(args.dir_icite)
        # append_outfile:
        #   * Append TOP, CIT, CLI, and REF into file if it is not already present
        #   * Print TOP to the screen
        if args.append_outfile:
            args.load_citations = True
            args.load_references = True
        return args

    # pylint: disable=too-many-arguments
    def run_icite(self, pmid2icitepaper_all, dnldr, args, argparser):
        """Run iCite/PubMed"""
        pmid2icitepaper_cur = self.run_icite_pre(pmid2icitepaper_all, args, argparser)
        if not pmid2icitepaper_cur:
            dnldr.prt_api_msgs()
            return
        # Write the report with citations and references for each PMID into its own file
        if args.O:
            self._wr_papers(pmid2icitepaper_cur, dnldr, args.print_header)
        # Write the succinct report each PMID to the screen
        else:
            self.run_icite_wr(pmid2icitepaper_cur, args, dnldr)

    def run_icite_pre(self, pmid2icitepaper_all, args, argparser):
        """Run iCite/PubMed"""
        ## print('ALL', pmid2icitepaper_all)
        if not pmid2icitepaper_all and not args.print_keys and not args.print_header:
            argparser.print_help()
            self._prt_infiles(args.infile)
        pmid2icitepaper_cur = {p: o for p, o in pmid2icitepaper_all.items() if o is not None}
        if not pmid2icitepaper_cur:
            # pylint: disable=line-too-long
            self._prt_no_icite(set(pmid2icitepaper_all.keys()).difference(pmid2icitepaper_cur.keys()))
            return {}
        return pmid2icitepaper_cur

    @staticmethod
    def _prt_infiles(infiles):
        """Print input files"""
        if infiles:
            for fin in infiles:
                print('**ERROR: NO PMIDs found in: {F}'.format(F=fin))

    @staticmethod
    def run_icite_wr(pmid2icitepaper, args, dnldr):
        """Print papers, including citation counts"""
        dct = get_outfile(args.outfile, args.append_outfile, args.force_write)
        if dct['outfile'] is None and not args.O:
            dnldr.prt_papers(pmid2icitepaper, prt=stdout)
        else:
            if args.verbose:
                dnldr.prt_papers(pmid2icitepaper, prt=stdout)
            elif args.append_outfile:
                # Print TOP to stdout if append_outfile
                for icitepaper in pmid2icitepaper.values():
                    icitepaper.prt_top(prt=stdout)
            if dct['outfile'] is not None:
                dnldr.wr_papers(dct['outfile'], pmid2icitepaper, dct['force_write'], dct['mode'])

    def _wr_papers(self, pmid2icitepaper, dnldr, print_header):
        """Write one icite report per PMID into dir_icite/PMID.txt"""
        for pmid, paper in pmid2icitepaper.items():
            fout_txt = self.cfg.get_fullname_icite('{PMID}.txt'.format(PMID=pmid))
            with open(fout_txt, 'w') as prt:
                if print_header:
                    prt_hdr(prt)
                dnldr.prt_papers({pmid:paper}, prt)
                print('  WROTE: {TXT}'.format(TXT=fout_txt))

    @staticmethod
    def _prt_no_icite(pmids):
        if not pmids:
            return
        print('**NOTE: No NIH iCite papers found for: {Ps}'.format(
            Ps=' '.join(str(p) for p in pmids)))


# Copyright (C) 2019-present DV Klopfenstein. All rights reserved.
